#---------------------------------------------------------------------------
# Metrics/losses 
# Mostly for segmentation
#---------------------------------------------------------------------------

import torch 
from torch import nn 
import torch.nn.functional as F
import numpy as np

#---------------------------------------------------------------------------
# Metrics base class

class Metric(nn.Module):
    """Abstract class for all metrics. Built to store the metric value, average, and name. In biom3d structure, metrics must have a `name`, a `self.reset` method that reset the stored values to zero and a `self.update` method that update the average value with the value stored in the `val` argument.
    
    To create a new metric, sub-class the Metric class, override the `self.__init__` and the `self.forward` methods. In the `self.__init__` set the `name` argument value and in `self.forward` set `val` argument value with the current value of your metric. 

    Parameters
    ----------
    name : str, default=None
        Name of the metric. Mainly for display purpose.
    """
    def __init__(self, name=None):
        super(Metric, self).__init__()
        self.name = name
        self.reset()

    def reset(self):
        self.val = 0
        self.avg = 0
        self.sum = 0
        self.count = 0
    
    def forward(self, preds, trues):
        pass

    def update(self, n=1):
        with torch.no_grad():
            self.sum += self.val * n
            self.count += n
            self.avg = self.sum / self.count
    
    def str(self):
        return str(self.val)
    
    def __str__(self):
        to_print = self.avg if self.avg!=0 else self.val
        return "{} {:.3f}".format(self.name, to_print)

#---------------------------------------------------------------------------
# Metrics/losses for semantic segmentation

class Dice(Metric):
    """Dice score computation. 

    Parameters
    ----------
    use_softmax : bool, default=False
        Whether the logit include the background for softmax computation. If True, then the background channel will be removed.
    dim : tuple, default=()
        The dimensions to which the score will be computed. If the default value is used, then the score will be computed for all dimension which might cause a class imbalance. To balance your channel, prefer setting `dim=(2,3,4)` if you deal with 3D images and `dim=(2,3)` for 2D images. 
    name : str, default=None
        Name of the metric. Mainly for display purpose.
    """
    def __init__(self, use_softmax=False, dim=(), name=None):
        super(Dice, self).__init__()
        self.name = name
        self.dim = dim
        self.use_softmax = use_softmax # if use softmax then remove bg

    def forward(self, inputs, targets, smooth=1):
        if self.use_softmax:
            # for dice computation, remove the background and flatten
            inputs = inputs.softmax(dim=1)

            if not all([i == j for i, j in zip(inputs.shape, targets.shape)]):
                # if this is not the case then gt is probably not already a one hot encoding
                targets_oh = torch.zeros(inputs.shape, device=inputs.device)
                targets_oh.scatter_(1, targets.long(), 1)
            else:
                targets_oh = targets

            # remove background
            inputs = inputs[:,1:]
            targets = targets_oh[:,1:]
        else:
            inputs = inputs.sigmoid()

        #flatten label and prediction tensors
        # inputs = inputs.reshape(-1)
        # targets = targets.reshape(-1)

        intersection = (inputs * targets).sum(dim=self.dim)                            
        dice = (2.*intersection + smooth)/(inputs.sum(dim=self.dim) + targets.sum(dim=self.dim) + smooth)  

        self.val = 1 - dice.mean() if self.training else dice.mean()
        return self.val  
    
class DiceBCE(Metric):
    """Dice-binary cross-entropy score computation. This metric can be considered as a loss function. 

    Parameters
    ----------
    use_softmax : bool, default=False
        Whether the logit include the background for softmax computation. If True, then the background channel will be removed.
    dim : tuple, default=()
        The dimensions to which the score will be computed. If the default value is used, then the score will be computed for all dimension which might cause a class imbalance. To balance your channel, prefer setting `dim=(2,3,4)` if you deal with 3D images and `dim=(2,3)` for 2D images. 
    name : str, default=None
        Name of the metric. Mainly for display purpose.
    """
    def __init__(self, use_softmax=False, dim=(), name=None):
        super(DiceBCE, self).__init__()
        self.use_softmax = use_softmax # if use softmax then remove bg for dice computation
        self.name = name
        self.dim = dim # axis defined for the dice score 
        self.bce = torch.nn.CrossEntropyLoss(reduction='mean')

    def forward(self, inputs, targets, smooth=1):
        #comment out if your model contains a sigmoid or equivalent activation layer
        if self.use_softmax:
            BCE = self.bce(inputs, targets.argmax(dim=1).long())

            # for dice computation, remove the background and flatten
            # inputs = inputs.softmax(dim=1)[:,1:].reshape(-1)
            # targets = targets[:,1:].reshape(-1)
            inputs = inputs.softmax(dim=1)

            if not all([i == j for i, j in zip(inputs.shape, targets.shape)]):
                # if this is not the case then gt is probably not already a one hot encoding
                targets_oh = torch.zeros(inputs.shape, device=inputs.device)
                targets_oh.scatter_(1, targets.long(), 1)
            else:
                targets_oh = targets

            # remove background
            inputs = inputs[:,1:]
            targets = targets_oh[:,1:]

        else:
            # keep the background and flatten
            # inputs = inputs.reshape(-1)
            # targets = targets.reshape(-1)
            BCE = F.binary_cross_entropy_with_logits(inputs, targets, reduction='mean')
            inputs = inputs.sigmoid()


        intersection = (inputs * targets).sum(dim=self.dim)                            
        dice_loss = 1 - (2.*intersection + smooth)/(inputs.sum(dim=self.dim) + targets.sum(dim=self.dim) + smooth)
        Dice_BCE = BCE + dice_loss.mean()
        
        self.val = Dice_BCE
        return self.val 
    
class IoU(Metric):
    """Intersection over union score computation. 

    Parameters
    ----------
    use_softmax : bool, default=False
        Whether the logit include the background for softmax computation. If True, then the background channel will be removed.
    dim : tuple, default=()
        The dimensions to which the score will be computed. If the default value is used, then the score will be computed for all dimension which might cause a class imbalance. To balance your channel, prefer setting `dim=(2,3,4)` if you deal with 3D images and `dim=(2,3)` for 2D images. 
    name : str, default=None
        Name of the metric. Mainly for display purpose.
    """
    def __init__(self, use_softmax=False, dim=(), name=None):
        super(IoU, self).__init__()
        self.use_softmax = use_softmax # if use softmax then remove bg
        self.dim = dim
        self.name = name 

    def forward(self, inputs, targets, smooth=1):
        if self.use_softmax:
            # inputs = inputs.softmax(dim=1)
            inputs = inputs.softmax(dim=1)

            if not all([i == j for i, j in zip(inputs.shape, targets.shape)]):
                # if this is not the case then gt is probably not already a one hot encoding
                targets_oh = torch.zeros(inputs.shape, device=inputs.device)
                targets_oh.scatter_(1, targets.long(), 1)
            else:
                targets_oh = targets

            # remove background
            inputs = inputs[:,1:]
            targets = targets_oh[:,1:]
        else:
            inputs = inputs.sigmoid()

        # inputs = inputs.reshape(-1)
        # targets = targets.reshape(-1)       
        
        #intersection is equivalent to True Positive count
        #union is the mutually inclusive area of all labels & predictions 
        intersection = (inputs * targets).sum(dim=self.dim)
        total = (inputs + targets).sum(dim=self.dim)
        union = total - intersection 
        
        iou = (intersection + smooth)/(union + smooth)
        
        self.val = 1 - iou.mean() if self.training else iou.mean()
        return self.val 

class MSE(Metric):
    """Mean Square Error loss.

    Parameters
    ----------
    name : str, default=None
        Name of the metric. Mainly for display purpose.
    """
    def __init__(self, name=None):
        super(MSE, self).__init__()
        self.name = name

    def forward(self, inputs, targets):

        self.val = torch.nn.functional.mse_loss(inputs, targets, reduction='mean')
        return self.val

class CrossEntropy(Metric):
    """Cross entropy loss.

    Parameters
    ----------
    name : str, default=None
        Name of the metric. Mainly for display purpose.
    """
    def __init__(self, name=None):
        super(CrossEntropy, self).__init__()
        self.name = name
        self.ce = torch.nn.CrossEntropyLoss(reduction='mean')

    def forward(self, inputs, targets):
        self.val = self.ce(inputs, targets)
        return self.val

#---------------------------------------------------------------------------
# Metric adaptation for deep supervision

class DeepMetric(Metric):
    """Deep supervision: a metric applied to several levels of the U-Net model. During the forward pass, this metric supposes that the inputs are a list of torch.Tensor that will be individually compared to the target.

    Parameters
    ----------
    metric : biom3d.metrics.Metric
        A metric that will be applied to each level of the U-Net model.
    alphas : list of float
        A list of coefficient applied to each feature map that starts with the deepest one. It must have a len of 6.
    name : str, default=None
        Name of the metric. Mainly for display purpose.
    metric_kwargs : dict, default={}
        A python dictionary that contains the keyword arguments required to define the metric.
    """
    def __init__(self,
        metric,
        alphas, # list of coefficient applied to each feature map, starts with the deepest one, must have a len of 6
        name=None,
        metric_kwargs={}):
        super(DeepMetric, self).__init__()
        self.metric = metric(**metric_kwargs)
        self.name = name 
        self.alphas = alphas 
    
    def forward(self, inputs, targets):
        """Forward pass.

        Parameters
        ----------
        inputs : list of torch.Tensor
            A list of batch of torch.Tensor that will be passed to the metric.
        targets : torch.Tensor
            A batch of torch.Tensor that will be compared to the inputs.
        """
        # inputs must be a list of network output
        # they are here all compared to the targets
        # the last inputs is supposed to be the final one
        self.val = 0
        for i in range(len(inputs)):
            if self.alphas[i]!=0:
                self.val += self.metric(inputs[i], targets)*self.alphas[i]
        return self.val

#---------------------------------------------------------------------------
# nnUNet metrics

def sum_tensor(inp, axes, keepdim=False):
    axes = np.unique(axes).astype(int)
    if keepdim:
        for ax in axes:
            inp = inp.sum(int(ax), keepdim=True)
    else:
        for ax in sorted(axes, reverse=True):
            inp = inp.sum(int(ax))
    return inp

def get_tp_fp_fn_tn(net_output, gt, axes=None, mask=None, square=False):
    """
    net_output must be (b, c, x, y(, z)))
    gt must be a label map (shape (b, 1, x, y(, z)) OR shape (b, x, y(, z))) or one hot encoding (b, c, x, y(, z))
    if mask is provided it must have shape (b, 1, x, y(, z)))
    :param net_output:
    :param gt:
    :param axes: can be (, ) = no summation
    :param mask: mask must be 1 for valid pixels and 0 for invalid pixels
    :param square: if True then fp, tp and fn will be squared before summation
    :return:
    """
    if axes is None:
        axes = tuple(range(2, len(net_output.size())))

    shp_x = net_output.shape
    shp_y = gt.shape

    with torch.no_grad():
        if len(shp_x) != len(shp_y):
            gt = gt.view((shp_y[0], 1, *shp_y[1:]))

        if all([i == j for i, j in zip(net_output.shape, gt.shape)]):
            # if this is the case then gt is probably already a one hot encoding
            y_onehot = gt
        else:
            gt = gt.long()
            y_onehot = torch.zeros(shp_x, device=net_output.device)
            y_onehot.scatter_(1, gt, 1)

    tp = net_output * y_onehot
    fp = net_output * (1 - y_onehot)
    fn = (1 - net_output) * y_onehot
    tn = (1 - net_output) * (1 - y_onehot)

    if mask is not None:
        tp = torch.stack(tuple(x_i * mask[:, 0] for x_i in torch.unbind(tp, dim=1)), dim=1)
        fp = torch.stack(tuple(x_i * mask[:, 0] for x_i in torch.unbind(fp, dim=1)), dim=1)
        fn = torch.stack(tuple(x_i * mask[:, 0] for x_i in torch.unbind(fn, dim=1)), dim=1)
        tn = torch.stack(tuple(x_i * mask[:, 0] for x_i in torch.unbind(tn, dim=1)), dim=1)

    if square:
        tp = tp ** 2
        fp = fp ** 2
        fn = fn ** 2
        tn = tn ** 2

    if len(axes) > 0:
        tp = sum_tensor(tp, axes, keepdim=False)
        fp = sum_tensor(fp, axes, keepdim=False)
        fn = sum_tensor(fn, axes, keepdim=False)
        tn = sum_tensor(tn, axes, keepdim=False)

    return tp, fp, fn, tn

class SoftDiceLoss(nn.Module):
    def __init__(self, apply_nonlin=None, batch_dice=False, do_bg=True, smooth=1.):
        """
        """
        super(SoftDiceLoss, self).__init__()

        self.do_bg = do_bg
        self.batch_dice = batch_dice
        self.apply_nonlin = apply_nonlin
        self.smooth = smooth

    def forward(self, x, y, loss_mask=None):
        shp_x = x.shape

        if self.batch_dice:
            axes = [0] + list(range(2, len(shp_x)))
        else:
            axes = list(range(2, len(shp_x)))

        if self.apply_nonlin is not None:
            x = self.apply_nonlin(x)

        tp, fp, fn, _ = get_tp_fp_fn_tn(x, y, axes, loss_mask, False)

        nominator = 2 * tp + self.smooth
        denominator = 2 * tp + fp + fn + self.smooth

        dc = nominator / (denominator + 1e-8)

        if not self.do_bg:
            if self.batch_dice:
                dc = dc[1:]
            else:
                dc = dc[:, 1:]
        dc = dc.mean()

        return -dc

class RobustCrossEntropyLoss(nn.CrossEntropyLoss):
    """
    this is just a compatibility layer because my target tensor is float and has an extra dimension
    """
    def forward(self, input, target):
        if len(target.shape) == len(input.shape):
            assert target.shape[1] == 1
            target = target[:, 0]
        return super().forward(input, target.long())

# class RobustCrossEntropyLoss(nn.NLLLoss):
#     def __init__(self):
#         super(RobustCrossEntropyLoss, self).__init__()
#         # self.log_softmax = torch.nn.LogSoftmax(dim=-1)
#         self.criterion = torch.nn.NLLLoss()

#     def forward(self, input, target):
#         if len(target.shape) == len(input.shape):
#             assert target.shape[1] == 1
#             target = target[:, 0]
#         # self.val = self.ce(inputs, targets)
#         # yhat = torch.sigmoid(inputs).clamp(min=1e-3, max=1-1e-3)
#         # print("metric: inputs shape and target", inputs.shape, targets)
#         # yhat = torch.log(inputs.softmax(dim=1).clamp(min=1e-3, max=1-1e-3))
#         yhat = torch.log(input.type(torch.float32).softmax(dim=1))
#         # y = F.one_hot(y, num_classes=self.num_class).float()
#         # self.val = -y*((1-yhat) ** self.gamma) * torch.log(yhat) - (1-y) * (yhat ** self.gamma) * torch.log(1-yhat)
#         # self.val = self.criterion(yhat, targets.long())
#         return super().forward(yhat, target.long())
#         # return self.val


class DC_and_CE_loss(Metric):
    def __init__(self, soft_dice_kwargs, ce_kwargs, aggregate="sum", square_dice=False, weight_ce=1, weight_dice=1,
                 log_dice=False, ignore_label=None, name=None):
        """
        CAREFUL. Weights for CE and Dice do not need to sum to one. You can set whatever you want.
        :param soft_dice_kwargs:
        :param ce_kwargs:
        :param aggregate:
        :param square_dice:
        :param weight_ce:
        :param weight_dice:
        """
        super(DC_and_CE_loss, self).__init__()
        if ignore_label is not None:
            assert not square_dice, 'not implemented'
            ce_kwargs['reduction'] = 'none'
        self.log_dice = log_dice
        self.weight_dice = weight_dice
        self.weight_ce = weight_ce
        self.aggregate = aggregate
        self.ce = RobustCrossEntropyLoss(**ce_kwargs)

        self.ignore_label = ignore_label

        if not square_dice:
            self.dc = SoftDiceLoss(apply_nonlin=lambda x: F.softmax(x, 1), **soft_dice_kwargs)
        
        self.name = name

    def forward(self, net_output, target):
        """
        target must be b, c, x, y(, z) with c=1
        :param net_output:
        :param target:
        :return:
        """
        
        if target.shape[1] != 1:
            target = target.argmax(dim=1).long().unsqueeze(dim=1)

        if self.ignore_label is not None:
            assert target.shape[1] == 1, 'not implemented for one hot encoding'
            mask = target != self.ignore_label
            target[~mask] = 0
            mask = mask.float()
        else:
            mask = None

        dc_loss = self.dc(net_output, target, loss_mask=mask) if self.weight_dice != 0 else 0
        if self.log_dice:
            dc_loss = -torch.log(-dc_loss)

        ce_loss = self.ce(net_output, target[:, 0].long()) if self.weight_ce != 0 else 0
        if self.ignore_label is not None:
            ce_loss *= mask[:, 0]
            ce_loss = ce_loss.sum() / mask.sum()

        if self.aggregate == "sum":
            result = self.weight_ce * ce_loss + self.weight_dice * dc_loss
        else:
            raise NotImplementedError("nah son") # reserved for other stuff (later)
        
        self.val = result
        return self.val

#---------------------------------------------------------------------------



#---------------------------------------------------------------------------
# Callback are periodically called during training. 
# There are currently 3 different periods:
# - training 
# - epoch
# - batch
#---------------------------------------------------------------------------

import torch
import os
# from telegram_send import send
import matplotlib.pyplot as plt
from torch.utils.tensorboard import SummaryWriter

import numpy as np

#----------------------------------------------------------------------------
# Base classes

class Callback(object):
    """Abstract base class used to build new callbacks.

    Callback are periodically called during training. 
    There are currently 3 different periods:
    - training 
    - epoch
    - batch
    Callbacks are called either before or after a period. 

    Each method starting by `on_` can be overridden. This method will be called at a certain time point during the training process.
    For instance, the `on_epoch_end` method will be called in the end of each epoch. 
    """

    def __init__(self):
        self.metrics = None

    def set_trainer(self, metrics):
        self.metrics = metrics

    def on_batch_begin(self, batch):
        pass

    def on_batch_end(self, batch):
        pass

    def on_epoch_begin(self, epoch):
        pass

    def on_epoch_end(self, epoch):
        pass

    def on_train_begin(self, epoch=None):
        pass

    def on_train_end(self):
        pass

# dict based callbacks

class Callbacks(Callback):
    """Child of biom3d.callbacks.Callback and used to compile all the callbacks together. 

    Parameters
    ----------
    callbacks : list of biom3d.callback.Callback
        A list of callbacks.
    """
    def __init__(self, callbacks):
        super().__init__()
        if isinstance(callbacks, Callbacks):
            callbacks = callbacks.callbacks
        self.callbacks = callbacks
        if callbacks is None:
            self.callbacks = {}
    
    def __getitem__(self, name): return self.callbacks[name]

    def set_trainer(self, trainer):
        for callback in self.callbacks.values():
            callback.set_trainer(trainer)

    def on_batch_begin(self, batch):
        for callback in self.callbacks.values():
            callback.on_batch_begin(batch)

    def on_batch_end(self, batch):
        for callback in self.callbacks.values():
            callback.on_batch_end(batch)

    def on_epoch_begin(self, epoch):
        for callback in self.callbacks.values():
            callback.on_epoch_begin(epoch)

    def on_epoch_end(self, epoch):
        for callback in self.callbacks.values():
            callback.on_epoch_end(epoch)

    def on_train_begin(self, epoch=None):
        for callback in self.callbacks.values():
            callback.on_train_begin()

    def on_train_end(self):
        for callback in self.callbacks.values():
            callback.on_train_end()

#----------------------------------------------------------------------------
# Savers            
            
class ModelSaver(Callback): # TODO: save best_only
    """The model saver saves the model, the epoch, the optimizer and the loss in the end of each epoch. It can also save the best model.
    
    Parameters
    ----------
    model : torch.nn.Module or a list of torch.nn.Module
        A torch module to store. If the model is a list then it is considered as been [student,teacher]
    optimizer : torch.optim.Optimizer
        A torch optimizer.
    path : str
        Name of the model to store. The `.pth` extension is automatically added and the best model is stored with `_best.pth` extension.
    every_epoch : int, default=2
        Period to save the model.
    save_best : bool, default=True
        Whether to save the best model.
    loss : torch.nn.Module
        Loss function.
    saved_loss : torch.nn.Module
        Loss saved alongside the model.
    """
    def __init__(self,
        model, 
        optimizer, 
        path='unet', 
        every_epoch=2, 
        save_best=True, 
        loss=None, # loss used for saving the best model, generally the val loss
        saved_loss=None, # loss being saved
        ):

        super().__init__()
        self.model = model
        self.optimizer = optimizer
        self.path_last = path + '.pth'
        self.path_best = path + '_best.pth'
        self.every_epoch = every_epoch
        self.save_best = save_best
        self.best_loss = float('inf')
        self.loss = loss
        self.saved_loss = saved_loss

    def on_epoch_end(self, epoch):
        if isinstance(self.model, list):
            save_dict = {
                'epoch':   epoch,
                'student': self.model[0].state_dict(), 
                'teacher': self.model[1].state_dict(), 
                'opt':     self.optimizer.state_dict(),
                'loss':    self.saved_loss.state_dict(),
                }
        else: 
            save_dict = {
                    'epoch':    epoch,
                    'model':    self.model.state_dict(), 
                    'opt':      self.optimizer.state_dict(),
                    'loss':     self.saved_loss.state_dict(),
                    }
        if epoch % self.every_epoch == 0:
            torch.save(save_dict, self.path_last)
            print('Save model to {}'.format(self.path_last))

        # save best model if needed 
        if self.save_best and (self.loss.avg < self.best_loss) and (self.loss.avg != 0):
            torch.save(save_dict, self.path_best)
            print('Save best model to {}'.format(self.path_best))
            self.best_loss = self.loss.avg
            
                

class LogSaver(Callback):
    """Save logs in a CSV file in the end of each epoch.

    This callback creates a `log.csv` file in the log folder.
    
    Parameters
    ----------
    log_dir : str
        Path to the folder where the CSV file will be stored. 
    train_loss : torch.nn.Module
        Training loss.
    val_loss : torch.nn.Module
        Validation loss.
    train_metrics : list of torch.nn.Module
        List of training metrics.
    val_metrics : list of torch.nn.Module
        List of validation metrics.
    scheduler : biom3d.callback.Callback
        Learning rate callback.
    """
    def __init__(self, 
        log_dir,            # path to where the csv file will be store
        train_loss,
        val_loss=None,
        train_metrics=None,
        val_metrics=None,
        scheduler=None,
        # save_best=False,    # save the best metrics in another text file ? 
        # every_batch=10
        ):    # save period (in batch)

        # old documentation about save best:
        # save_best : bool, default=True
        #     Whether to save the best loss in other CSV file. This will create another CSV file that will be called `log_best.csv` in the log folder.
        # every_batch : int, default=10
        #     Batch period when to save the log.

        self.path = os.path.join(log_dir,'log.csv')

        self.crt_epoch = 1

        self.train_loss = train_loss
        self.val_loss = val_loss
        self.train_metrics = train_metrics
        self.val_metrics = val_metrics

        self.scheduler = scheduler if hasattr(scheduler, 'get_last_lr') else None

        # if save_best:
        #     self.save_best = save_best # self.tracked_loss = val_loss
        #     self.best_loss = float('inf')
        #     self.best_path = log_dir + '/log_best.csv'
        #     self.f_best = None

        # self.every_batch = every_batch

        f = open(self.path, "a")
        if os.stat(self.path).st_size == 0: # if the file is empty
            self.write_file_head()
        f.close()

    def write_file_head(self):
        """Write the head of the CSV file.
        """
        # write the head of the log file
        f = open(self.path, "a")
        head = "epoch"
        if self.scheduler is not None:
            head += ',learning_rate'
        head += ",train_loss"
        if self.val_loss is not None:
            head += ",val_loss"
        if self.train_metrics is not None:
            for m in self.train_metrics:
                head += "," + m.name
        if self.val_metrics is not None:
            for m in self.val_metrics:
                head += "," + m.name
        f.write(head + "\n")

    # def on_epoch_begin(self, epoch):
    #     self.crt_epoch = epoch+1

    # def on_batch_end(self, batch):
    #     if batch % self.every_batch == 0:
    #         template =  str(self.crt_epoch)
    #         template += "," + str(batch)
    #         template += "," + str(self.scheduler.get_last_lr()[0])
    #         template += "," + str(self.train_loss.val.item()) # TODO: save the avg value only
    #         # BUG fix below: TODO, simplify it! 
    #         val_loss = self.val_loss.val if type(self.val_loss.val)==int else self.val_loss.val.item()
    #         template += "," + str(val_loss)
    #         for m in self.train_metrics: template += "," + str(m.val.item())
    #         for m in self.val_metrics: 
    #             # TODO: cf. previous comment
    #             val_m = m.avg if type(m.avg)==int else m.avg.item() 
    #             template += "," + str(val_m)
    #         self.f.write(template + "\n")
        
            # save best if needed
            # if self.save_best \
            #     and (self.best_loss > self.val_loss.val) \
            #     and (self.val_loss.avg != 0): # to avoid saving during the first epoch
            #     self.best_loss = self.val_loss.avg
            #     self.f_best = open(self.best_path, "w") # open it in write mode
            #     self.write_file_head(self.f_best)
            #     self.f_best.write(template)
            #     self.f_best.close()
    
    def on_epoch_end(self, epoch):
        f = open(self.path, "a")
        template =  str(epoch)

        # add the scheduler value
        if self.scheduler is not None:
            template += "," + str(self.scheduler.get_last_lr()[0])

        # add the learning rate and the training loss
        template += "," + str(self.train_loss.avg.item()) # TODO: save the avg value only
        
        # add the validation loss if needed
        if self.val_loss is not None:
            val_loss = self.val_loss.avg if type(self.val_loss.avg)==float else self.val_loss.avg.item()
            template += "," + str(val_loss)
        
        # add the training metrics
        if self.train_metrics is not None:
            for m in self.train_metrics: template += "," + str(m.val.item())

        # adde the validation metrics
        if self.val_metrics is not None:
            for m in self.val_metrics: 
                val_m = m.avg if type(m.avg)==int else m.avg.item() 
                template += "," + str(val_m)
        
        # write in the output file
        f.write(template + "\n")
        f.close()

class ImageSaver(Callback):
    """The image saver callback saves a small snapshot of the raw image, the prediction and the ground truth. By default, it uses the validation dataloader to load a batch of 3D images, makes a prediction with the model, then uses the first images of the batch and the prediction and, to display a 2D image, uses the first channel of each image (input, prediction, ground truth).

    Parameters
    ----------
    image_dir : str
        Path to the image snapshot folder, where the snapshots will be stored.
    model : torch.nn.Module
        A torch module that will be used to make prediction.
    val_dataloader : torch.utils.data.Dataloader or iter
        An iterator used to generate a batch of input image and mask.
    use_sigmoid : bool, default=True
        Whether to use a sigmoid or a softmax activation on the model predictions.
    every_epoch : int, default=1
        Epoch period when the image snapshot will be stored.
    plot_size : int, default=1
        Number of images to plot in the snapshot. 
    use_fp16 : bool, default=True
        Whether the model has been trained with AMP or not.
    """
    def __init__(self, 
        image_dir, 
        model,
        val_dataloader,
        use_sigmoid=True,
        every_epoch=1,
        plot_size=1,
        use_fp16=True,
        ):
        self.image_dir = image_dir
        self.every_epoch = every_epoch
        self.model = model 
        self.use_sigmoid = use_sigmoid
        self.val_dataloader = val_dataloader
        self.plot_size = plot_size
        self.use_fp16 = use_fp16

    def on_epoch_end(self, epoch):
        if epoch % self.every_epoch == 0:
            self.model.eval()
            with torch.no_grad():
                for i in range(self.plot_size):
                # make prediction
                    X, y = next(iter(self.val_dataloader))
                    # X, y = self.val_dataloader.get_sample()
                    if torch.cuda.is_available():
                        X, y = X.cuda(), y.cuda()
                    with torch.cuda.amp.autocast(self.use_fp16):
                        pred = self.model(X)
                        if type(pred)==list:
                            pred = pred[-1]
                    if self.use_sigmoid:
                        pred = (torch.sigmoid(pred)>0.5).int()*255
                    else: 
                        pred = (pred.softmax(dim=1).argmax(dim=1).unsqueeze(dim=1)).int()*255
                    l = [X, y, pred.detach()]
                    for j in range(len(l)):
                        _,_,channel,_,_ = l[j].shape
                        l[j] = l[j][-1, -1, channel//2, ...].cpu().numpy().astype(float)
                    X, y, pred = l

                    # plot 
                    # plt.figure(figsize=(15,5))
                    plt.figure(dpi=100)
                    # print original
                    plt.subplot(self.plot_size,3,3*i+1)
                    plt.imshow(X)
                    plt.title('raw')
                    plt.axis('off')
                    # print prediction
                    plt.subplot(self.plot_size,3,3*i+2)
                    plt.imshow(pred)
                    plt.title('pred')
                    plt.axis('off')
                    # print ground truth
                    plt.subplot(self.plot_size,3,3*i+3)
                    plt.imshow(y)
                    plt.title('ground_truth')
                    plt.axis('off')
                    del X, y, pred
                im_path = os.path.join(self.image_dir,'image_' + str(epoch) + '.png')
                print("Save image to {}".format(im_path))
                plt.savefig(im_path)
                plt.close()

class TensorboardSaver(Callback):
    """The tensorboard callback is used to plot the loss and metrics in the tensorboard interface. To start tensorboard execute the following command in a terminal opened in the biom3d project folder: `tensorboard --logdir=logs/`.
    
    This callback labels the loss with `Loss/train` and `Loss/test` and the metrics with `Metrics/name_of_your_metric`.

    Parameters
    ----------
    log_dir : str
        Path to the log folder where the curve will be stored.
    train_loss : torch.nn.Module
        Training loss function.
    val_loss : torch.nn.Module
        Validation loss function.
    train_metrics : list of torch.nn.Module
        List of the training metrics.
    val_metrics : list of torch.nn.Module
        List of the validation metrics.
    batch_size : int
        Size of the minibatch. Is used to compute the number of iteration so that the x-axis of the curve is the number of image and not the number of batch.
    n_batch_per_epoch : int
        Number of the batch per epoch. Is used to compute the number of iteration so that the x-axis of the curve is the number of image and not the number of batch.
    """
    def __init__(self, 
        log_dir, 
        train_loss, 
        val_loss, 
        train_metrics, 
        val_metrics, 
        batch_size,
        n_batch_per_epoch):

        self.writer = SummaryWriter(log_dir=log_dir)
        self.train_loss = train_loss
        self.val_loss = val_loss
        self.train_metrics = train_metrics
        self.val_metrics = val_metrics
        self.batch_size = batch_size
        self.n_batch_per_epoch = n_batch_per_epoch

        self.crt_epoch = 0
    
    def on_epoch_begin(self, epoch):
        self.crt_epoch = epoch
    
    # def on_batch_end(self, batch):
    def on_epoch_end(self, epoch):
        # n_iter = (self.n_batch_per_epoch * self.crt_epoch + batch) * self.batch_size
        n_iter = (epoch+1) * self.batch_size * self.n_batch_per_epoch
        self.writer.add_scalar('Loss/train', self.train_loss.avg, n_iter)
        if self.val_loss:
            self.writer.add_scalar('Loss/test', self.val_loss.avg, n_iter)
        if self.train_metrics:
            for m in self.train_metrics: self.writer.add_scalar('Metrics/'+m.name,m.avg,n_iter)
        if self.val_metrics:
            for m in self.val_metrics: self.writer.add_scalar('Metrics/'+m.name,m.avg,n_iter)

#----------------------------------------------------------------------------
# Printer

class LogPrinter(Callback):
    """Print the log in the terminal prompt.

    Parameters
    ----------
    metrics : list of torch.nn.Module
        List of the metrics to display. The loss can be amoung them.
    nbof_eopchs : int
        Number of epochs in the training.
    nbof_batchs : int
        Number of batches per epoch.
    every_batch : int, default=10
        Batch period when to print the log.
    """
    def __init__(
        self,
        metrics,
        nbof_epochs,
        nbof_batches,
        every_batch=10):

        self.nbof_epochs = nbof_epochs
        self.nbof_batches = nbof_batches

        self.metrics = metrics

        self.every_batch = every_batch

    def on_epoch_begin(self, epoch):
        print("Epoch [{:>3d}/{:>3d}]".format(epoch+1, self.nbof_epochs))

    def on_batch_end(self, batch):
        if batch % self.every_batch == 0:
            template = "Batch [{:>3d}/{:>3d}]".format(batch, self.nbof_batches)
            for i in range(len(self.metrics)):
                template += ", {}".format(self.metrics[i])
            print(template)

# class Telegram(Callback):
#     """
#     Send message when training finishes
#     """
#     def __init__(self, loss, test_loss=None):
#         self.loss = loss
#         self.test_loss = test_loss
#         self.template = "Finished training with loss {} and test loss {}"
    
#     def on_train_end(self):
#         send(messages=[self.template.format(self.loss.val, self.test_loss.avg)])

#----------------------------------------------------------------------------
# schedulers

class LRSchedulerMultiStep(Callback):
    """Multi-step learning rate scheduler. Uses torch.optim.lr_scheduler.MultiStepLR

    For more details, see : https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.MultiStepLR
    
    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        Training optimizer.
    milestones : list of int
        List of epoch when to decay the learning rate. Must be increasing. 
    gamma : float, default=0.1
        Multiplicative factor of the learning rate decay. 
    """
    def __init__(self, optimizer, milestones, gamma=0.1):
        self.scheduler = torch.optim.lr_scheduler.MultiStepLR(optimizer, milestones, gamma=gamma, verbose=False)

    def get_last_lr(self):
        """Getter for the last learning rate.
        """
        return self.scheduler.get_last_lr()
    
    def on_epoch_end(self, epoch):
        self.scheduler.step()
        print("Current learning rate: {}".format(self.scheduler.get_last_lr()))

class LRSchedulerCosine(Callback):
    """Cosine learning rate scheduler. Uses torch.optim.lr_scheduler.CosineAnnealingLR

    For more details, see : https://pytorch.org/docs/stable/generated/torch.optim.lr_scheduler.CosineAnnealingLR
    
    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        Training optimizer.
    T_max : int
        Maximum number of iteration. By default, you can set this to the number of epochs. 
    """
    def __init__(self, optimizer, T_max):
        self.scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(optimizer, T_max, eta_min=1e-6, verbose=False)

    def get_last_lr(self):
        """Getter for the last learning rate.
        """
        return self.scheduler.get_last_lr()
    
    def on_epoch_end(self, epoch):
        self.scheduler.step()
        print("Current learning rate: {}".format(self.scheduler.get_last_lr()))

class LRSchedulerPoly(Callback):
    """Polynomial scheduler. Similar to nnU-Net learning rate scheduler.

    .. math::
        \\begin{aligned}
            lr_{new} = lr_{initial} * (1 - \\frac{epoch_{current}}{epoch_{max}})^{exponent}
        \\end{aligned}

    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        Training optimizer.
    initial_lr : float
        Initial learning rate.
    max_epochs : int
        Number of epochs in the training.
    exponent : float, default=0.9
        Must be between 0 and 1. Exponent of the polynomial decay.
    """
    def __init__(self, optimizer, initial_lr, max_epochs, exponent=0.9):
        self.initial_lr = initial_lr
        self.max_epochs = max_epochs
        self.exponent = exponent
        self.optimizer = optimizer

    # def get_last_lr(self):
    #     return self.optimizer.param_groups[0]['lr']
    
    # def on_train_begin(self, epoch=None):
    #     self.optimizer.param_groups[0]['lr'] = self.initial_lr * (1 - epoch / self.max_epochs)**self.exponent
    #     print("Current learning rate: {}".format(self.optimizer.param_groups[0]['lr']))

    def on_epoch_begin(self, epoch):
        self.optimizer.param_groups[0]['lr'] = self.initial_lr * (1 - epoch / self.max_epochs)**self.exponent
        print("Current learning rate: {}".format(self.optimizer.param_groups[0]['lr']))


# class LRSchedulerCosine(Callback):
#     """
#     Multi-step scheduler only for now
#     """
#     def __init__(self, optimizer, T_max):
#         self.scheduler = torch.optim.lr_scheduler.CosineAnnealingWarmRestarts(optimizer, T_0=50, T_mult=2, eta_min=1e-6, last_epoch=-1, verbose=False)

#     def get_last_lr(self):
#         return self.scheduler.get_last_lr()
    
#     def on_epoch_end(self, epoch):
#         self.scheduler.step()
#         print("Current learning rate: {}".format(self.scheduler.get_last_lr()))

class ForceFGScheduler(Callback):
    """Force foreground scheduler. Uses a polynomial scheduler.
    We force the model to "see" the foreground more often in the beginning of the training and progressively reduce the foreground rate.
    The callback calls the dataloader.dataset.set_fg_rate method. The dataset must thus possess a set_fg_rate function.
    This function does not change directly the way the foreground images are presented to the deep learning model. This should be implemented in the Dataset class.

    Parameters
    ----------
    dataloader : torch.utils.data.Dataloader 
        A dataloader which dataset member possesses a `set_fg_rate` method.
    initial_rate : float 
        Initial foreground rate. If the rate is equal to 1, it means that we only present foreground. On the contrary, if the foreground rate is set to 0, then the patch will be random. (Again, this depends on how the foreground forcing is defined in the dataloader.dataset).
    min_rate : float
        Minimal/final foreground rate. Can be set to zero but it is not recommended. A good choice could be 0.33 (such as in nnU-Net implementation).
    max_epochs : int
        Number of epochs in the training.
    exponent : float, default=0.9
        Must be between 0 and 1. Exponent of the polynomial decay.
    """
    def __init__(self, dataloader, initial_rate, min_rate, max_epochs, exponent=0.9):
        self.dataloader = dataloader
        self.initial_rate = initial_rate
        self.min_rate = min_rate
        self.max_epochs = max_epochs
        self.exponent = exponent

    # def on_train_begin(self, epoch=None):
    #     self.dataloader.dataset.set_fg_rate(self.initial_rate)
    #     print("Current foreground rate: {}".format(self.initial_rate))
    
    def on_epoch_begin(self, epoch):
        crt_rate = (self.initial_rate-self.min_rate) * (1 - epoch / self.max_epochs)**self.exponent + self.min_rate
        self.dataloader.dataset.set_fg_rate(crt_rate) 
        print("Current foreground rate: {}".format(crt_rate))

class OverlapScheduler(Callback):
    """Overlapping rate scheduler. Uses a polynomial scheduler.
    We progressively reduce the minimum overlap between the global patches and the local patches.
    An overlap of 1 means that the local patch is completely included in the global patch. An overlap of 0 or less, means that the local patch can be located outside of the global patch. The exact definition of the relationship between the local and the global patch should be defined the dataset.
    
    Parameters
    ----------
    dataloader : torch.utils.data.Dataloader 
        A dataloader which dataset member possesses a `set_min_overlap` method.
    initial_rate : float 
        Initial overlapping rate. 
    min_rate : float
        Minimal/final overlapping rate. 
    max_epochs : int
        Number of epochs in the training.
    exponent : float, default=0.9
        Must be between 0 and 1. Exponent of the polynomial decay.
    """
    def __init__(self, dataloader, initial_rate, min_rate, max_epochs, exponent=0.9):
        self.dataloader = dataloader
        self.initial_rate = initial_rate
        self.min_rate = min_rate
        self.max_epochs = max_epochs
        self.exponent = exponent

    # def on_train_begin(self, epoch=None):
        self.dataloader.dataset.set_min_overlap(self.initial_rate)
        print("Current overlap: {}".format(self.initial_rate))
    
    def on_epoch_begin(self, epoch):
        crt_rate = (self.initial_rate-self.min_rate) * (1 - epoch / self.max_epochs)**self.exponent + self.min_rate
        self.dataloader.dataset.set_min_overlap(crt_rate)
        print("Current overlap: {}".format(crt_rate))

class GlobalScaleScheduler(Callback):
    """Global scale scheduler. Uses a polynomial scheduler.
    We progressively reduce the scale of the global_crop from image size to patch/local_crop size.

    Parameters
    ----------
    dataloader : torch.utils.data.Dataloader 
        A dataloader which dataset member possesses a `set_global_crop` method.
    initial_rate : float 
        Initial global scale rate. 
    min_rate : float
        Minimal/final global scale rate. 
    max_epochs : int
        Number of epochs in the training.
    exponent : float, default=0.9
        Must be between 0 and 1. Exponent of the polynomial decay.
    """
    def __init__(self, dataloader, initial_rate, min_rate, max_epochs, exponent=0.9):
        self.dataloader = dataloader
        self.initial_rate = initial_rate
        self.min_rate = min_rate
        self.max_epochs = max_epochs
        self.exponent = exponent

    # def on_train_begin(self, epoch=None):
    #     self.dataloader.dataset.set_global_crop(self.initial_rate)
    #     print("Current global crop scale: {}".format(self.initial_rate))
    
    def on_epoch_begin(self, epoch):
        crt_rate = (self.initial_rate-self.min_rate) * (1 - epoch / self.max_epochs)**self.exponent + self.min_rate
        self.dataloader.dataset.set_global_crop(crt_rate)
        print("Current global crop scale: {}".format(crt_rate))

class WeightDecayScheduler(Callback):
    """Weight decay scheduler. Used for DINO re-implementation.
    
    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        Training optimizer.
    initial_wd : float
        Initial weight decay.
    final_wd : float
        Final weight decay.
    nb_epochs : int
        Number of epochs in the training.
    use_poly : bool, default=True
        Whether to use polynomial scheduler instead of the default cosine scheduler.
    exponent : float, default=0.9
        Must be between 0 and 1. Exponent of the polynomial decay.
    """
    def __init__(self, optimizer, initial_wd, final_wd, nb_epochs, use_poly=False, exponent=0.9):
        self.optimizer = optimizer
        self.initial_wd = initial_wd
        self.final_wd = final_wd
        self.nb_epochs = nb_epochs
        self.use_poly = use_poly 
        self.exponent = exponent

    # def on_train_begin(self, epoch=None):
    #     self.optimizer.param_groups[0]["weight_decay"] = self.initial_wd
    #     print("Current weight decay:", self.optimizer.param_groups[0]["weight_decay"] )
    
    def on_epoch_begin(self, epoch):
        if self.use_poly:
            self.optimizer.param_groups[0]["weight_decay"] = self.final_wd + (self.initial_wd - self.final_wd) * (1 - epoch / self.nb_epochs)**self.exponent
        else: 
            self.optimizer.param_groups[0]["weight_decay"] = self.final_wd + 0.5 * (self.initial_wd - self.final_wd) * (1 + np.cos(np.pi * epoch / self.nb_epochs))
        print("Current weight decay:", self.optimizer.param_groups[0]["weight_decay"] )

class MomentumScheduler(Callback):
    """Momentum scheduler. Used for DINO re-implementation.
    
    Parameters
    ----------
    optimizer : torch.optim.Optimizer
        Training optimizer.
    initial_momentum : float
        Initial momentum decay.
    final_momentum : float
        Final momentum decay.
    nb_epochs : int
        Number of epochs in the training.
    mode : str, default='poly'
        Scheduling mode. Can be one of: 'poly', 'exp', 'linear' or None. If None, use cosine scheduler.
    exponent : float, default=0.9
        Must be between 0 and 1. Exponent of the polynomial decay.
    """
    def __init__(self, initial_momentum, final_momentum, nb_epochs, mode='poly', exponent=0.9):
        self.initial_momentum = initial_momentum
        self.final_momentum = final_momentum
        self.nb_epochs = nb_epochs
        self.crt_momentum = None
        self.mode = mode 
        self.exponent = exponent
    
    def __getitem__(self, epoch):
        if self.mode=='poly':
            self.crt_momentum = self.final_momentum + (self.initial_momentum - self.final_momentum) * (1 - epoch / self.nb_epochs)**self.exponent
        elif self.mode=='exp':
            self.crt_momentum = self.final_momentum + (self.initial_momentum - self.final_momentum) * (np.exp(-self.exponent*epoch / self.nb_epochs))
        elif self.mode=='linear':
            if epoch > self.nb_epochs:
                self.crt_momentum = self.final_momentum
            else:
                self.crt_momentum = self.initial_momentum + (self.final_momentum - self.initial_momentum) * epoch / self.nb_epochs
        # elif self.mode=='mix': # linear increase and then an exponential increase
        #     if epoch > self.nb_epochs:
        #         self.crt_momentum = self.final_momentum
        #     else:
        #         self.crt_momentum = self.initial_momentum + (self.final_momentum - self.initial_momentum) * epoch / self.nb_epochs
        else: # cosine
            self.crt_momentum = self.final_momentum + 0.5 * (self.initial_momentum - self.final_momentum) * (1 + np.cos(np.pi * epoch / self.nb_epochs))
        return self.crt_momentum

    def on_epoch_end(self, epoch):
        print("Current teacher momentum:", self.crt_momentum)

class DatasetSizeScheduler(Callback):
    """Dataset size scheduler. 
    We progressively increase the size of the dataset to help the Arcface training. NO PROOF THAT IT IMPROVES THE FINAL PERFORMANCE.
    The dataloader must have a dataset that has a `set_num_classes` method.
    At each epoch the size of the dataset is incremented by 1.

    Parameters
    ----------
    dataloader : torch.utils.data.Dataloader
        Torch dataloader. The dataloader must have a dataset that has a `set_num_classes` method.
    model : torch.nn.Module
        Torch module that must have a `set_num_classes` method.
    max_dataset_size : int
        Maximum size of the dataset.
    min_dataset_size : int
        Minimum size of the dataset. Used in the beginning of the training.
    """
    def __init__(self, dataloader, model, max_dataset_size, min_dataset_size=5):
        self.dataloader = dataloader
        self.model = model
        self.max_dataset_size = max_dataset_size
        self.min_dataset_size = min_dataset_size

    def on_epoch_begin(self, epoch):
        dataset_size = min(epoch+self.min_dataset_size, self.max_dataset_size)
        self.dataloader.dataset.set_dataset_size(dataset_size)
        self.model.set_num_classes(dataset_size)
        print("Current dataset size: {}".format(dataset_size))

#----------------------------------------------------------------------------
# metrics updater

class MetricsUpdater(Callback):
    """Update the metrics averages by calling the `update` method of each metric.

    Parameters
    ----------
    metrics : list of torch.nn.Module
        List of metrics and losses to update.
    batch_size : int
        Batch size.
    """
    def __init__(self, metrics, batch_size):
        self.metrics = metrics
        self.batch_size = batch_size

    def on_epoch_begin(self, epoch=None):
        for m in self.metrics: m.reset()

    def on_batch_end(self, batch=None):
        for m in self.metrics: m.update(self.batch_size)

#----------------------------------------------------------------------------


#---------------------------------------------------------------------------
# Dataset primitives for 3D segmentation dataset
# solution: patch approach with the whole dataset into memory 
#---------------------------------------------------------------------------

import os
import pickle
import numpy as np 
import torchio as tio
import random 
from torch.utils.data import Dataset
# from monai.data import CacheDataset
import pandas as pd 
# from tifffile import imread

from biom3d.utils import centered_pad, get_folds_train_test_df, adaptive_imread

#---------------------------------------------------------------------------
# utilities to random crops

def centered_crop(img, msk, center, crop_shape, margin=np.zeros(3)):
    """Do a crop, forcing the location voxel to be located in the center of the crop.
    """
    img_shape = np.array(img.shape)[1:]
    center = np.array(center)
    assert np.all(center>=0) and np.all(center<img_shape), "[Error] Center must be located inside the image. Center: {}, Image shape: {}".format(center, img_shape)
    crop_shape = np.array(crop_shape)
    margin = np.array(margin)
    
    # middle of the crop
    # start = np.maximum(0,center-crop_shape//2+margin).astype(int)
    start = (center-crop_shape//2+margin).astype(int)

    # assert that the end will not be out of the crop
    # start = start - np.maximum(start+crop_shape-img_shape, 0)

    # end = center+(crop_shape-crop_shape//2)
    end = start+crop_shape

    # we make sure that we are not out of the image shape
    start = np.maximum(0,start)
    
    idx = [slice(0,img.shape[0])]+list(slice(s[0], s[1]) for s in zip(start, end))
    idx = tuple(idx)
    
    crop_img = img[idx]
    crop_msk = msk[idx]
    return crop_img, crop_msk

def located_crop(img, msk, location, crop_shape, margin=np.zeros(3)):
    """Do a crop, forcing the location voxel to be located in the crop.
    """
    img_shape = np.array(img.shape)[1:]
    location = np.array(location)
    crop_shape = np.array(crop_shape)
    margin = np.array(margin)
    
    lower_bound = np.maximum(0,location-crop_shape+margin)
    higher_bound = np.maximum(lower_bound+1,np.minimum(location-margin, img_shape-crop_shape))
    start = np.random.randint(low=lower_bound, high=np.maximum(lower_bound+1,higher_bound))
    end = start+crop_shape
    
    idx = [slice(0,img.shape[0])]+list(slice(s[0], s[1]) for s in zip(start, end))
    idx = tuple(idx)
    
    crop_img = img[idx]
    crop_msk = msk[idx]
    return crop_img, crop_msk

def foreground_crop(img, msk, final_size, fg_margin, fg=None, use_softmax=True):
    """Do a foreground crop.
    """
    if fg is not None and len(list(fg.keys()))>0:
        locations = fg[random.choice(list(fg.keys()))]
    else:
        if tuple(msk.shape)[0]==1:
            # then we consider that we don't have a one hot encoded label
            rnd_label = random.randint(1,msk.max() if msk.max()>0 else 1)
            locations = np.argwhere(msk[0] == rnd_label)
        else:
            # then we have a one hot encoded label
            rnd_label = random.randint(int(use_softmax),tuple(msk.shape)[0]-1)
            locations = np.argwhere(msk[rnd_label] == 1)

    if np.array(locations).size==0: # bug fix when having empty arrays 
        img, msk = random_crop(img, msk, final_size, force_in=False)
    else:
        center=random.choice(locations) # choose a random voxel of this label
        img, msk = centered_crop(img, msk, center, final_size, fg_margin)
    return img, msk

def centered_pad(img, final_size, msk=None):
    """
    centered pad an img and msk to fit the final_size
    """
    final_size = np.array(final_size)
    img_shape = np.array(img.shape[1:])
    
    start = (final_size-np.array(img_shape))//2
    start = start * (start > 0)
    end = final_size-(img_shape+start)
    end = end * (end > 0)
    
    pad = np.append([[0,0]], np.stack((start,end),axis=1), axis=0)
    pad_img = np.pad(img, pad, 'constant', constant_values=0)
    
    if msk is not None:
        pad_msk = np.pad(msk, pad, 'constant', constant_values=0)
        return pad_img, pad_msk
    else: 
        return pad_img

def random_crop(img, msk, crop_shape, force_in=True):
    """
    randomly crop a portion of size prop of the original image size.
    """ 
    img_shape = np.array(img.shape)[1:]
    assert len(img_shape)==len(crop_shape),"[Error] Not the same dimensions! Image shape {}, Crop shape {}".format(img_shape, crop_shape)
    
    if force_in: # force the crop to be located in image shape range
        start = np.random.randint(0, np.maximum(1,img_shape-crop_shape))
        end = start+crop_shape
        
        idx = [slice(0,img.shape[0])]+list(slice(s[0], s[1]) for s in zip(start, end))
        idx = tuple(idx)
        
        crop_img = img[idx]
        crop_msk = msk[idx]
    else: # the crop will be chosen randomly and then padded if needed
        # the crop might be too small but will be padded with zeros
        start = np.random.randint(0, img_shape)
        crop_img, crop_msk = centered_crop(img=img, msk=msk, center=start, crop_shape=crop_shape)
        # pad if needed
        if np.any(np.array(crop_img.shape)[1:]-crop_shape)!=0:
            crop_img, crop_msk = centered_pad(img=crop_img, msk=crop_msk, final_size=crop_shape)

    return crop_img, crop_msk

def random_crop_pad(img, msk, final_size, fg_rate=0.33, fg_margin=np.zeros(3), fg=None, use_softmax=True):
    """
    random crop and pad if needed.
    """
    if type(img)==list: # then batch mode
        imgs, msks = [], []
        for i in range(len(img)):
            img_, msk_ = random_crop_pad(img[i], msk[i], final_size)
            imgs += [img_]
            msks += [msk_]
        return np.array(imgs), np.array(msks) # can convert to array as they should have now all the same shape
    
    # choose if using foreground centrered or random alignement
    force_fg = random.random()
    if fg_rate>0 and force_fg<fg_rate:
        img, msk = foreground_crop(img, msk, final_size, fg_margin, fg=fg, use_softmax=use_softmax)
    else:
        # or random crop
        img, msk = random_crop(img, msk, final_size, force_in=False)
        
    # pad if needed
    if np.any(np.array(img.shape)[1:]-final_size)!=0:
        img, msk = centered_pad(img=img, msk=msk, final_size=final_size)
    return img, msk

def random_crop_resize(img, msk, crop_scale, final_size, fg_rate=0.33, fg_margin=np.zeros(3)):
    """
    random crop and resize if needed.
    Args:
    crop_scale: >=1
    """
    final_size = np.array(final_size)
        
    # determine crop shape
    max_crop_shape = np.round(final_size * crop_scale).astype(int)
    crop_shape = np.random.randint(final_size, max_crop_shape+1)
    
    # choose if using foreground centrered or random alignement
    force_fg = random.random()
    if fg_rate>0 and force_fg<fg_rate:
        rnd_label = random.randint(0,msk.shape[0]-1) # choose a random label
        
        locations = np.argwhere(msk[rnd_label] == 1)
        
        if locations.size==0: # bug fix when having empty arrays 
            img, msk = random_crop(img, msk, crop_shape)
        else:
            center=random.choice(locations) # choose a random voxel of this label
            img, msk = located_crop(img, msk, center, crop_shape, fg_margin)
    else:
        # or random crop
        img, msk = random_crop(img, msk, crop_shape)
    
    # resize if needed
    if np.any(np.array(img.shape)[1:]-final_size)>0:
        sub = tio.Subject(img=tio.ScalarImage(tensor=img), msk=tio.LabelMap(tensor=msk))
        sub = tio.Resize(final_size)(sub)
        img, msk = sub.img.tensor, sub.msk.tensor
    elif np.any(np.array(img.shape)[1:]-final_size)<0:
        img, msk = centered_pad(img, msk, final_size)
    return img, msk

#---------------------------------------------------------------------------

class LabelToLong:
    def __init__(self, label_name):
        self.label_name = label_name
        
    def __call__(self, subject):
        if self.label_name in subject.keys():
            subject[self.label_name].set_data(subject[self.label_name].data.long())
        return subject

#---------------------------------------------------------------------------

class SemSeg3DPatchFast(Dataset):
    """
    with DataLoader
    """
    def __init__(
        self,
        img_dir,
        msk_dir,
        batch_size, 
        patch_size,
        nbof_steps,
        folds_csv  = None, 
        fold       = 0, 
        val_split  = 0.25,
        train      = True,
        use_aug    = True,
        aug_patch_size = None,
        fg_dir  = None,
        fg_rate = 0.33, # if > 0, force the use of foreground, needs to run some pre-computations (note: better use the foreground scheduler)
        crop_scale = 1.0, # if > 1, then use random_crop_resize instead of random_crop_pad
        load_data = False, # if True, loads the all dataset into computer memory (faster but more memory expensive)
        use_softmax = True,
        ):

        self.img_dir = img_dir
        self.msk_dir = msk_dir
        self.fg_dir = fg_dir

        self.batch_size = batch_size
        self.patch_size = patch_size
        self.aug_patch_size = aug_patch_size

        self.nbof_steps = nbof_steps

        self.load_data = load_data
        
        # get the training and validation names 
        if folds_csv is not None:
            df = pd.read_csv(folds_csv)
            trainset, testset = get_folds_train_test_df(df, verbose=False)

            self.fold = fold
            
            self.val_imgs = trainset[self.fold]
            del trainset[self.fold]
            self.train_imgs = []
            for i in trainset: self.train_imgs += i

        else: 
            all_set = sorted(os.listdir(img_dir))
            assert len(all_set) > 0, "[Error] Incorrect path for folder of images or your folder is empty."
            np.random.shuffle(all_set) # shuffle all_set
            val_split = np.round(val_split * len(all_set)).astype(int)
            if val_split == 0: val_split=1
            self.train_imgs = all_set[val_split:]
            self.val_imgs = all_set[:val_split]
            testset = []
        
        self.train = train
        if self.train:
            print("current fold: {}\n \
            length of the training set: {}\n \
            length of the validation set: {}\n \
            length of the testing set: {}".format(fold, len(self.train_imgs), len(self.val_imgs), len(testset)))

        self.fnames = self.train_imgs if self.train else self.val_imgs

        # print train and validation image names
        print("{} images: {}".format("Training" if self.train else "Validation", self.fnames))
        
        if self.load_data:
            print("Loading the whole dataset into computer memory...")
            def load_data(fnames, fg_dir=None):
                imgs_data = []
                msks_data = []
                fg_data   = []
                for idx in range(len(fnames)):
                    # file names
                    img_path = os.path.join(self.img_dir, fnames[idx])
                    msk_path = os.path.join(self.msk_dir, fnames[idx])

                    # load img and msks
                    imgs_data += [adaptive_imread(img_path)[0]]
                    msks_data += [adaptive_imread(msk_path)[0]]

                    # load foreground 
                    if fg_dir is not None:
                        fg_path = os.path.join(self.fg_dir, fnames[idx][:fnames[idx].find(".")]+'.pkl')
                        fg = pickle.load(open(fg_path, 'rb'))
                        fg_data += [fg]
                return imgs_data, msks_data, fg_data

            self.imgs_data, self.msks_data, self.fg_data = load_data(self.fnames, fg_dir=fg_dir)
            print("Done!")

        self.use_aug = use_aug

        if self.use_aug:
            ps = np.array(self.patch_size)

            # [aug] flipping probabilities
            flip_prop=ps.min()/ps
            flip_prop/=flip_prop.sum()

            # [aug] 'axes' for tio.RandomAnisotropy
            anisotropy_axes=tuple(np.arange(len(ps))[ps/ps.min()>3].tolist())
            if len(anisotropy_axes)==0:
                anisotropy_axes=tuple(np.arange(len(ps)).tolist())

            # [aug] 'degrees' for tio.RandomAffine
            if np.any(ps/ps.min()>3): # then use dummy_2d
                # norm = ps*3/ps.min()
                # softmax=np.exp(norm)/sum(np.exp(norm))
                # degrees=softmax.min()*90/softmax
                # degrees = 180*ps.min()/ps
                degrees = tuple(180 if p==ps.argmin() else 0 for p in range(len(ps)))
            else:
                degrees = (-45,45)
                # degrees = 90

            # [aug] 'cropping'
            # the affine transform is computed on bigger patches than the other transform
            # that's why we need to crop the patch after potential affine augmentation
            start = (np.array(self.aug_patch_size)-np.array(self.patch_size))//2
            end = self.aug_patch_size-(np.array(self.patch_size)+start)
            cropping = (start[0],end[0],start[1],end[1],start[2],end[2])
            
            # the foreground-crop-function forces the foreground to be in the center of the patch
            # so that, when doing the second centrering crop, the foreground is still present in the patch,
            # that's why there is a margin here
            # [DEPRECATED] 2023/05/22 Guillaume: this is not the case anymore, will be removed
            # self.fg_margin = start 
            self.fg_margin = np.zeros(len(patch_size))


            # self.transform = tio.Compose([
            #     # spatial augmentations
            #     # tio.RandomFlip(p=0.8, axes=(0), flip_probability=flip_prop[0]),
            #     # tio.RandomFlip(p=0.8, axes=(1), flip_probability=flip_prop[1]),
            #     # tio.RandomFlip(p=0.8, axes=(2), flip_probability=flip_prop[2]),
            #     # tio.RandomFlip(p=0.2, axes=(0,1,2)),
            #     # tio.RandomAnisotropy(p=0.25, axes=anisotropy_axes, downsampling=(1,2)),
            #     tio.RandomAffine(p=0.25, scales=(0.7,1.4), degrees=degrees, translation=0),
            #     tio.Crop(p=1, cropping=cropping),

            #     tio.RandomFlip(p=0.7, axes=(0,1,2)),
            #     # tio.RandomElasticDeformation(p=0.2, num_control_points=4, locked_borders=1),
            #     # tio.OneOf({
            #     #     tio.RandomAffine(scales=0.1, degrees=10, translation=0): 0.8,
            #     #     tio.RandomElasticDeformation(): 0.2,
            #     # }),
                

            #     # intensity augmentations
            #     # tio.RandomMotion(p=0.2),
            #     # tio.RandomGhosting(p=0.2),
            #     # tio.RandomSpike(p=0.15),
            #     tio.RandomBiasField(p=0.15, coefficients=0.2),
            #     tio.RandomBlur(p=0.2, std=(0.5,1.5)),
            #     tio.RandomNoise(p=0.2, std=(0,0.1)),
            #     # tio.RandomSwap(p=0.2, patch_size=ps//8),
            #     tio.RandomGamma(p=0.3, log_gamma=(-0.35,0.4)),
            # ])
            # separate rotation as it requires a bigger patch size
            self.rotate = tio.Compose([
                                 tio.RandomAffine(scales=0, degrees=degrees, translation=0, default_pad_value=0),
                                 tio.Crop(cropping=cropping),
                                 LabelToLong(label_name='msk')])
            self.transform = tio.Compose([
                # pre-cropping to aug_patch_size
                # tio.OneOf({
                #     tio.Compose([
                #                  tio.RandomAffine(scales=0, degrees=degrees, translation=0, default_pad_value=0),
                #                  tio.Crop(cropping=cropping),
                #                  LabelToLong(label_name='msk')
                #                 ]): 0.2,
                #     tio.Crop(cropping=cropping): 0.8,
                # }),

                tio.Compose([tio.RandomAffine(scales=(0.7,1.4), degrees=0, translation=0),
                             LabelToLong(label_name='msk')
                            ], p=0.2),

                # spatial augmentations
                tio.RandomAnisotropy(p=0.1, axes=anisotropy_axes, downsampling=(1,1.5)),
                # tio.RandomAffine(p=0.25, scales=(0.7,1.4), degrees=degrees, translation=0),
                # tio.Crop(cropping=cropping),
                tio.RandomFlip(p=1, axes=(0,1,2)),
                # tio.OneOf({
                #     tio.RandomAffine(scales=0.1, degrees=10, translation=0): 0.8,
                #     tio.RandomElasticDeformation(): 0.2,
                # }),

                # intensity augmentations
                # tio.RandomMotion(p=0.2),
                # tio.RandomGhosting(p=0.2),
                # tio.RandomSpike(p=0.15),
                tio.RandomBiasField(p=0.15, coefficients=0.2),
                tio.RandomBlur(p=0.2, std=(0.5,1)),
                tio.RandomNoise(p=0.2, std=(0,0.1)),
                tio.RandomSwap(p=0.2, patch_size=ps//8),
                tio.RandomGamma(p=0.3, log_gamma=(-0.35,0.4)),
                # LabelToFloat(label_name='msk')
            ])

        # self.fg_rate = fg_rate if self.train else 1
        self.fg_rate = fg_rate
        self.crop_scale = crop_scale 
        assert self.crop_scale >= 1, "[Error] crop_scale must be higher or equal to 1"

        self.use_softmax = use_softmax
        self.batch_idx = 0
    
    def set_fg_rate(self,value):
        """
        setter function for the foreground rate class parameter
        """
        self.fg_rate = value

    def _do_fg(self):
        """
        determines whether to force the foreground depending on the batch idx
        """
        return not self.batch_idx < round(self.batch_size * (1 - self.fg_rate))
    
    def _update_batch_idx(self):
        self.batch_idx += 1
        if self.batch_idx >= self.batch_size:
            self.batch_idx = 0
    
    def __len__(self):
        # return len(self.train_imgs) if self.training else len(self.val_imgs)
        return self.nbof_steps*self.batch_size
    
    def __getitem__(self, idx):

        if self.load_data:
            img = self.imgs_data[idx%len(self.imgs_data)]
            msk = self.msks_data[idx%len(self.msks_data)]
            if len(self.fg_data)>0: fg = self.fg_data[idx%len(self.fg_data)]
            else: fg = None
        else:
            # img_fname = np.random.choice(fnames)
            img_fname = self.fnames[idx%len(self.fnames)]
            
            # file names
            img_path = os.path.join(self.img_dir, img_fname)
            msk_path = os.path.join(self.msk_dir, img_fname)

            # read the images
            img = adaptive_imread(img_path)[0]
            msk = adaptive_imread(msk_path)[0]

            # read foreground data
            if self.fg_dir is not None:
                fg_path = os.path.join(self.fg_dir, img_fname[:img_fname.find(".")]+'.pkl')
                fg = pickle.load(open(fg_path, 'rb'))
            else:
                fg = None

        # random crop and pad
        # rotation augmentation requires a larger patch size
        do_rot = random.random() < 0.2 # rotation proba = 0.2 
        final_size = self.aug_patch_size if self.use_aug and do_rot else self.patch_size
        fg_margin = self.fg_margin if self.use_aug else np.zeros(3)
        if self.train and self.crop_scale > 1:
            img, msk = random_crop_resize(
                img,
                msk,
                crop_scale=self.crop_scale,
                final_size=final_size,
                fg_rate=self.fg_rate,
                fg_margin=fg_margin,
                )
        else:
            img, msk = random_crop_pad(
                img,
                msk,
                final_size=final_size,
                # fg_rate=self.fg_rate,
                fg_rate=int(self._do_fg()),
                fg_margin=fg_margin,
                fg = fg,
                use_softmax=self.use_softmax
                )
            self._update_batch_idx()

        # data augmentation
        if self.use_aug:
            sub = tio.Subject(img=tio.ScalarImage(tensor=img), msk=tio.LabelMap(tensor=msk))
            if do_rot: sub = self.rotate(sub)
            sub = self.transform(sub)
            img, msk = sub.img.tensor, sub.msk.tensor
        
            # to float for msk
            msk = msk.float()
        else:
            # convert mask to float for validation
            msk = msk.astype(float)
        
        return img, msk

#---------------------------------------------------------------------------



#---------------------------------------------------------------------------
# 3D Resnet adapted from:
# https://github.com/akamaster/pytorch_resnet_cifar10 
#---------------------------------------------------------------------------

import torch.nn as nn
import torch.nn.functional as F
import numpy as np

from biom3d.utils import convert_num_pools

#---------------------------------------------------------------------------
# 3D Resnet encoder

__all__ = ['ResNet', 'resnet20', 'resnet32', 'resnet44', 'resnet56', 'resnet110', 'resnet1202']

def _weights_init(m):
    # classname = m.__class__.__name__
    #print(classname)
    # if isinstance(m, nn.Linear) or isinstance(m, nn.Conv3d):
    #     init.kaiming_normal_(m.weight)
    # if isinstance(m, nn.Conv3d):
    #     nn.init.xavier_normal_(m.weight)
    # elif isinstance(m, nn.BatchNorm3d) or isinstance(m, nn.BatchNorm1d):
    #     nn.init.constant_(m.weight, 1)
    #     nn.init.constant_(m.bias, 0)
    # elif isinstance(m, nn.Linear):
    #     nn.init.xavier_normal_(m.weight)
    #     if m.bias is not None:
    #         nn.init.constant_(m.bias, 0)
    if isinstance(m, nn.Conv2d):
        nn.init.kaiming_normal_(m.weight, mode="fan_out", nonlinearity="relu")
        if m.bias is not None:
            nn.init.constant_(m.bias, 0)
    elif isinstance(m, nn.BatchNorm2d):
        nn.init.constant_(m.weight, 1)
        nn.init.constant_(m.bias, 0)
    elif isinstance(m, nn.Linear):
        nn.init.normal_(m.weight, 0, 0.01)
        if m.bias is not None:
            nn.init.constant_(m.bias, 0)

class LambdaLayer(nn.Module):
    def __init__(self, lambd):
        super(LambdaLayer, self).__init__()
        self.lambd = lambd

    def forward(self, x):
        return self.lambd(x)

class GlobalAvgPool3d(nn.Module):
    def __init__(self):
        super(GlobalAvgPool3d, self).__init__()

    def forward(self,x):
        out = x.mean(dim=(-3,-2,-1))
        return out

class SmallEncoderBlock(nn.Module):
    def __init__(self, in_planes, planes, stride=1, option='B', is_last=False):
        super(SmallEncoderBlock, self).__init__()
        self.conv1 = nn.Conv3d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.is_last = is_last
        if not is_last:
            self.bn1 = nn.InstanceNorm3d(planes, affine=True)

    def forward(self, x):
        if not self.is_last:
            out = F.leaky_relu(self.bn1(self.conv1(x)), inplace=True)
        else:
            out = self.conv1(x)
        return out

class EncoderBlock(nn.Module):
    def __init__(self, in_planes, planes, stride=1, option='B', is_last=False):
        super(EncoderBlock, self).__init__()

        self.conv1 = nn.Conv3d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.InstanceNorm3d(planes, affine=True)
        self.conv2 = nn.Conv3d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)

        self.is_last = is_last
        if not is_last:
            self.bn2 = nn.InstanceNorm3d(planes, affine=True)

    def forward(self, x):
        out = F.leaky_relu(self.bn1(self.conv1(x)), inplace=True)
        if not self.is_last:
            out = F.leaky_relu(self.bn2(self.conv2(out)), inplace=True)
        else:
            out = self.conv2(out)
        return out

class VGGEncoder(nn.Module):
    def __init__(self, 
        block, 
        num_pools, 
        factor = 32,
        first_stride=[1,1,1], # the stride of the first layer convolution
        flip_strides = False, # whether to invert strides order. Flipped strides creates larger feature maps.
        use_emb=False, # use the embedding output (along with the existing ones)
        emb_dim=320,
        use_head=False,
        patch_size = None, # only needed when using the head
        in_planes = 1,
        ): 
        super(VGGEncoder, self).__init__()
        factors = [factor * i for i in [1,2,4,8,10,10,10]]
        self.in_planes = in_planes
        self.use_emb=use_emb
        self.use_head = use_head

        # computes the strides
        # for example: convert [3,5,5] into [[1 1 1],[1 2 2],[2 2 2],[2 2 2],[2 2 2],[1 2 2]]
        # max_pool = max(num_pools)
        # strides = []
        # for i in range(len(num_pools)):
        #     st = np.ones(max_pool)
        #     num_zeros = max_pool-num_pools[i]
        #     for j in range(num_zeros):
        #         st[j]=0
        #     st=np.roll(st,-num_zeros//2)
        #     strides += [st]
        # strides = np.array(strides).astype(int).T+1
        strides = convert_num_pools(num_pools=num_pools)
        if flip_strides: strides = np.flip(strides, axis=0)
        strides = np.vstack(([first_stride],strides))
        strides = strides.tolist()
        

        # computes the strides (old)
        # strides = [first_stride]
        # num_pools_ = np.array(num_pools)
        # while num_pools_.sum()!=0:
        #     stride = (num_pools_>0).astype(int)
        #     num_pools_ -= stride
        #     strides += [list(stride+1)]

        # defines the network
        self.layers = []
        for i in range(max(num_pools)+1):
            self.layers += [self._make_layer(block, factors[i], num_blocks=1, stride=strides[i], is_last=(i==max(num_pools) and use_emb))]
        
        self.layers = nn.ModuleList(self.layers)

        if use_emb:
            # factor = factors[max(num_pools)]
            # emb_size = factor # size of the embedding
            # self.global_pool=GlobalAvgPool3d()
            # self.global_pool=torch.nn.AdaptiveAvgPool3d((1,1,1))
            # self.global_pool=torch.nn.AdaptiveAvgPool3d((5,5,5))
            # self.fc = nn.Linear(factor, emb_dim, bias=True)
            # self.bn = nn.BatchNorm1d(emb_dim)
            # self.bn.weight.requires_grad = False
            if use_head:
                strides_ = (np.array(strides)).prod(axis=0)
                in_dim = (np.array(patch_size)/strides_).prod().astype(int)*in_planes*factors[-1]
                last_layer = nn.utils.weight_norm(nn.Linear(256, emb_dim, bias=False))
                # norm last layer
                last_layer.weight_g.data.fill_(1)
                last_layer.weight_g.requires_grad = False

                self.head = nn.Sequential(
                    nn.Linear(in_dim, 2048),
                    nn.GELU(),
                    nn.Dropout(p=0.5),
                    nn.Linear(2048, 2048),
                    nn.GELU(),
                    nn.Dropout(p=0.5),
                    nn.Linear(2048, 256), # bottleneck
                    last_layer,
                )

            

        self.apply(_weights_init)

    def _make_layer(self, block, planes, num_blocks, stride, is_last=False):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers += [block(self.in_planes, planes, stride, is_last=is_last)]
            self.in_planes = planes

        return nn.Sequential(*layers)

    def forward(self, x, use_encoder=False):
        # stores the intermediate outputs
        out = []
        for i in range(len(self.layers)):
            inputs = x if i==0 else out[-1]
            out += [self.layers[i](inputs)]
            
        if self.use_emb:
            # out = self.last_layer(out[-1])
            # out = self.global_pool(out[-1])
            # out = torch.flatten(out, 1)
            out = out[-1].view(out[-1].size(0), -1)
            # out = nn.functional.normalize(out, dim=-1, p=2)
            # if self.use_head:
            #     out = nn.functional.normalize(out, dim=-1, p=2)
            #     out = self.fc(out)
            if use_encoder:
                out = self.head(out)
            
        return out
#---------------------------------------------------------------------------




#---------------------------------------------------------------------------
# model predictors
# these functions returns the model predictions for a single input 
# TODO: re-structure with classes maybe?
#---------------------------------------------------------------------------

import torch 
import torchio as tio
import numpy as np
from skimage.io import imread
from tqdm import tqdm
# from scipy.ndimage.filters import gaussian_filter

from biom3d.utils import keep_biggest_volume_centered, adaptive_imread, resize_3d, keep_big_volumes

#---------------------------------------------------------------------------
# model predictor for segmentation

def load_img_seg(fname):
    img = imread(fname)

    # normalize
    img = (img - img.min()) / (img.max() - img.min())
    
    # to tensor
    img = torch.tensor(img, dtype=torch.float)

    # expand dim
    img = torch.unsqueeze(img, dim=0)
    img = torch.unsqueeze(img, dim=0)

    return img

def seg_predict(
    img_path,
    model,
    return_logit=False,
    ):
    """
    for one image path, load the image, compute the model prediction, return the prediction
    """
    img = load_img_seg(img_path)
    model.eval()
    with torch.no_grad():
        if torch.cuda.is_available():
            img = img.cuda()
        logit = model(img)[0][0]

    if return_logit: return logit.cpu().detach().numpy()

    out = (torch.sigmoid(logit)>0.5).int()*255
    return out.cpu().detach().numpy()

def seg_predict_old(img, model, return_logit=False):
    model.eval()
    with torch.no_grad():
        if torch.cuda.is_available():
            img = img.cuda()
        img = torch.unsqueeze(img, 0)
        logit = model(img)[0]

    if return_logit: return logit.cpu().detach().numpy()

    out = (torch.sigmoid(logit)>0.5).int()*255
    return out.cpu().detach().numpy()

#---------------------------------------------------------------------------
# model predictor for segmentation with patches

class LoadImgPatch:
    def __init__(
        self,
        fname,
        patch_size=(64,64,32),
        median_spacing=[],
        clipping_bounds=[],
        intensity_moments=[],
        ):
        
        self.fname = fname
        self.patch_size = np.array(patch_size)
        self.median_spacing = np.array(median_spacing)
        self.clipping_bounds = clipping_bounds
        self.intensity_moments = intensity_moments
        
        # prepare image
        # load the image
        img,metadata = adaptive_imread(self.fname)
        self.spacing = None if not 'spacing' in metadata.keys() else metadata['spacing']

        # store img shape (for post processing)
        self.img_shape = img.shape
        print("image shape: ",self.img_shape)
        
        # expand image dim
        if len(img.shape)==3:
            img = np.expand_dims(img, 0)
        elif len(img.shape)==4:
            # we consider as the channel dimension, the smallest dimension
            # it should be either the first or the last dim
            # if it is the last dim, then we move it to the first
            if np.argmin(img.shape)==3:
                img = np.moveaxis(img, -1, 0)
            elif np.argmin(img.shape)!=0:
                print("[Error] Invalid image shape:", img.shape)
        else:
            print("[Error] Invalid image shape:", img.shape)
    
        # preprocessing: resampling, clipping, z-normalization
        # resampling if needed
        if len(self.median_spacing)>0:
            resample = (self.median_spacing/self.spacing)[::-1] # transpose the dimension
            if resample.sum() > 0.1: # otherwise no need of spacing 
                sub = tio.Subject(img=tio.ScalarImage(tensor=img))
                sub = tio.Resample(resample)(sub)
                img = sub.img.numpy()
                print("Resampling required! From {} to {}".format(self.img_shape, img.shape))

        # clipping if needed
        if len(self.clipping_bounds)>0:
            img = np.clip(img, self.clipping_bounds[0], self.clipping_bounds[1])

        # normalize 
        if len(self.intensity_moments)>0:
            img = (img-self.intensity_moments[0])/self.intensity_moments[1]
        else:
            img = (img-img.mean())/img.std()

        # convert to tensor of float
        self.img = torch.from_numpy(img).float()
        
    def get_gridsampler(self):
        """
        Prepare image for model prediction and return a tio.data.GridSampler
        """
        # define the grid sampler 
        patch_overlap = np.maximum(self.patch_size//2, self.patch_size-np.array(self.img_shape))
        patch_overlap = np.ceil(patch_overlap/2).astype(int)*2
        sub = tio.Subject(img=tio.ScalarImage(tensor=self.img))
        sampler= tio.data.GridSampler(subject=sub, 
                                patch_size=self.patch_size, 
                                patch_overlap=patch_overlap,
                                padding_mode='constant')
        return sampler
    
    def post_process(self, logit):
        """
        resampling back the image after model prediction
        """
        if len(self.median_spacing)==0:
            return logit 
        
        resample = (self.spacing/self.median_spacing)[::-1] # transpose the dimension
        if resample.sum() > 0.1:
            sub = tio.Subject(img=tio.ScalarImage(tensor=logit))
            sub = tio.Resample(resample)(sub)
            
            # if the image has not the right size still, then try to crop it first
            if not np.array_equal(sub.img.shape[1:], self.img_shape):
                img = sub.img.tensor
                x,y,z = self.img_shape
                img = img[:,:x,:y,:z]
                
            # if the image has still not the right size still, then resize it
            if not np.array_equal(sub.img.shape[1:], self.img_shape):
                sub = tio.Resize(self.img_shape)(sub)
                img = sub.img.tensor
            return img
        else:
            return logit

def seg_predict_patch(
    img_path,
    model,
    return_logit=False,
    patch_size=None,
    tta=False,          # test time augmentation 
    median_spacing=[],
    clipping_bounds=[],
    intensity_moments=[],
    use_softmax=False,
    force_softmax=False,
    num_workers=4,
    enable_autocast=True, 
    keep_biggest_only=False,
    ):
    """
    for one image path, load the image, compute the model prediction, return the prediction
    """

    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    enable_autocast = torch.cuda.is_available() and enable_autocast # tmp, autocast seems to work only with gpu for now... 
    print('AMP {}'.format('enabled' if enable_autocast else 'disabled'))

    img_loader = LoadImgPatch(
        img_path,
        patch_size,
        median_spacing,
        clipping_bounds,
        intensity_moments,
    )

    img = img_loader.get_gridsampler()

    model.eval()
    with torch.no_grad():
        pred_aggr = tio.inference.GridAggregator(img, overlap_mode='hann')
        patch_loader = torch.utils.data.DataLoader(
            img, 
            batch_size=2, 
            drop_last=False, 
            shuffle  =False, 
            num_workers=num_workers, 
            pin_memory =True)

        for patch in tqdm(patch_loader):
            X = patch['img'][tio.DATA]
            if torch.cuda.is_available():
                X = X.cuda()
            
            if tta: # test time augmentation: flip around each axis
                with torch.autocast(device, enabled=enable_autocast):
                    pred=model(X).cpu()
                
                # flipping tta
                dims = [[2],[3],[4]]
                for i in range(len(dims)):
                    X_flip = torch.flip(X,dims=dims[i])

                    with torch.autocast(device, enabled=enable_autocast):
                        pred_flip = model(X_flip)
                        pred += torch.flip(pred_flip, dims=dims[i]).cpu()
                    
                    del X_flip, pred_flip
                
                pred = pred/(len(dims)+1)
                del X
            else:
                with torch.autocast(device, enabled=enable_autocast):
                    pred=model(X).cpu()
                    del X

            pred_aggr.add_batch(pred, patch[tio.LOCATION])
        
        print("Prediction done!")

        print("Aggregation...")
        logit = pred_aggr.get_output_tensor().float()
        print("Aggregation done!")
    
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    # post-processing:
    print("Post-processing...")
    logit = img_loader.post_process(logit)

    if return_logit: 
        print("Post-processing done!")
        return logit

    if use_softmax:
        out = (logit.softmax(dim=0).argmax(dim=0)).int()
    elif force_softmax:
        # if the training has been done with a sigmoid activation and we want to export a softmax
        # it is possible to use `force_softmax` argument
        sigmoid = (logit.sigmoid()>0.5).int()
        softmax = (logit.softmax(dim=0).argmax(dim=0)).int()+1
        cond = sigmoid.max(dim=0).values
        out = torch.where(cond>0, softmax, 0)
    else:
        out = (logit.sigmoid()>0.5).int()
    out = out.numpy()

    # TODO: the function below is too slow
    if keep_biggest_only:
        if len(out.shape)==3:
            out = keep_biggest_volume_centered(out)
        elif len(out.shape)==4:
            tmp = []
            for i in range(out.shape[0]):
                tmp += [keep_biggest_volume_centered(out[i])]
            out = np.array(tmp)

    out = out.astype(np.byte) 
    print("Post-processing done!")
    print("Output shape:",out.shape)
    return out

#---------------------------------------------------------------------------
# new predictor

def seg_predict_patch_2(
    img,
    original_shape,
    model,
    conserve_size=False, # force the logit to be the same size as the input
    patch_size=None,
    tta=False,          # test time augmentation 
    num_workers=4,
    enable_autocast=True, 
    use_softmax=True,   # DEPRECATED!
    keep_biggest_only=False, # DEPRECATED!
    **kwargs, # just for handling other image metadata
    ):
    """
    for one image, compute the model prediction, return the predicted logit
    """
    device = 'cuda' if torch.cuda.is_available() else 'cpu'
    enable_autocast = torch.cuda.is_available() and enable_autocast # tmp, autocast seems to work only with gpu for now... 
    print('AMP {}'.format('enabled' if enable_autocast else 'disabled'))

    # make original_shape 3D
    original_shape = original_shape[-3:]

    # get grid sampler
    overlap = 0.5
    patch_size = np.array(patch_size)

    # check that overlap is smaller or equal to image shape
    patch_overlap = np.maximum(patch_size*overlap, patch_size-np.array(img.shape[-3:]))

    # round patch_overlap
    patch_overlap = (np.ceil(patch_overlap*overlap)/overlap).astype(int)

    # if patch_overlap is equal to one of the patch_size dimension then torchio throw an error
    patch_overlap = np.minimum(patch_overlap, patch_size-1)

    sub = tio.Subject(img=tio.ScalarImage(tensor=img))
    sampler= tio.data.GridSampler(subject=sub, 
                            patch_size=patch_size, 
                            patch_overlap=patch_overlap,
                            padding_mode='constant')

    model.to(device).eval()

    with torch.no_grad():
        pred_aggr = tio.inference.GridAggregator(sampler, overlap_mode='hann')
        patch_loader = torch.utils.data.DataLoader(
            sampler, 
            batch_size=2, 
            drop_last=False, 
            shuffle  =False, 
            num_workers=num_workers, 
            pin_memory =True)

        for patch in tqdm(patch_loader):
            X = patch['img'][tio.DATA]
            if torch.cuda.is_available():
                X = X.cuda()
            
            if tta: # test time augmentation: flip around each axis
                with torch.autocast(device, enabled=enable_autocast):
                    pred=model(X).cpu()
                
                # flipping tta
                dims = [[2],[3],[4],[3,2],[4,2],[4,3],[4,3,2]]
                for i in range(len(dims)):
                    X_flip = torch.flip(X,dims=dims[i])

                    with torch.autocast(device, enabled=enable_autocast):
                        pred_flip = model(X_flip)
                        pred += torch.flip(pred_flip, dims=dims[i]).cpu()
                    
                    del X_flip, pred_flip
                
                pred = pred/(len(dims)+1)
                del X
            else:
                with torch.autocast(device, enabled=enable_autocast):
                    pred=model(X).cpu()
                    del X

            pred_aggr.add_batch(pred, patch[tio.LOCATION])
        
        print("Prediction done!")

        print("Aggregation...")
        logit = pred_aggr.get_output_tensor().float()
        print("Aggregation done!")
    
    model.cpu()
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

    # reshape the logit so it has the same size as the original image
    if conserve_size:
        return resize_3d(logit, original_shape, order=3)

    return logit

def seg_postprocessing(
        logit,
        original_shape=None,
        use_softmax=True,
        force_softmax=False,
        keep_big_only=False,
        keep_biggest_only=False,
        return_logit=False,
        **kwargs, # just for handling other image metadata
    ):
    """
    Post-process the logit (model output) to obtain the final segmentation mask. Can optionally remove some noise. 
    """
    # make original_shape 3D
    original_shape = original_shape[-3:]

    # post-processing:
    print("Post-processing...")

    if return_logit: 
        if original_shape is not None:
            logit = resize_3d(logit, original_shape, order=3)
        print("Post-processing done!")
        return logit

    if use_softmax:
        out = (logit.softmax(dim=0).argmax(dim=0)).int().numpy()        
    elif force_softmax:
        # if the training has been done with a sigmoid activation and we want to export a softmax
        # it is possible to use `force_softmax` argument
        sigmoid = (logit.sigmoid()>0.5).int()
        softmax = (logit.softmax(dim=0).argmax(dim=0)).int()+1
        cond = sigmoid.max(dim=0).values
        out = torch.where(cond>0, softmax, 0).numpy()        
    else:
        out = (logit.sigmoid()>0.5).int().numpy()
        
    # resampling
    if original_shape is not None:
        if use_softmax or force_softmax:
            out = resize_3d(np.expand_dims(out,0), original_shape, order=1, is_msk=True).squeeze()
        else: 
            out = resize_3d(out, original_shape, order=1, is_msk=True)
    
    if keep_big_only and keep_biggest_only:
        print("[Warning] Incompatible options 'keep_big_only' and 'keep_biggest_only' have both been set to True. Please deactivate one! We consider here only 'keep_biggest_only'.")
    # TODO: the function below is too slow
    if keep_biggest_only or keep_big_only:
        fct = keep_biggest_volume_centered if keep_biggest_only else keep_big_volumes
        if len(out.shape)==3:
            out = fct(out)
        elif len(out.shape)==4:
            tmp = []
            for i in range(out.shape[0]):
                tmp += [fct(out[i])]
            out = np.array(tmp)

    out = out.astype(np.uint8)    
    
    print("Post-processing done!")
    print("Output shape:",out.shape)
    return out

#---------------------------------------------------------------------------



#---------------------------------------------------------------------------
# Predictions with Omero
# This script can download data from Omero, compute predictions,
# and upload back into Omero.
#---------------------------------------------------------------------------

import argparse
import os
from omero.cli import cli_login
from omero.gateway import BlitzGateway

from biom3d import omero_downloader 
try:
    from biom3d import omero_uploader
except:
    pass
from biom3d import pred  

def run(obj, target, log, dir_out, host=None, user=None, pwd=None, upload_id=None, ext="_predictions"):
    print("Start dataset/project downloading...")
    if host is not None:
        datasets, dir_in = omero_downloader.download_object(user, pwd, host, obj, target)
    else:
        with cli_login() as cli:
            datasets, dir_in = omero_downloader.download_object_cli(cli, obj, target)

    print("Done downloading dataset/project!")

    print("Start prediction...")
    if 'Dataset' in obj:
        dir_in = os.path.join(dir_in, datasets[0].name)
        dir_out = os.path.join(dir_out, datasets[0].name)
        if not os.path.isdir(dir_out):
            os.makedirs(dir_out, exist_ok=True)
        dir_out = pred.pred(log, dir_in, dir_out)

        # eventually upload the dataset back into Omero [DEPRECATED]
        if upload_id is not None and host is not None:

            # create a new Omero Dataset
            dataset_name = os.path.basename(dir_in)
            if len(dataset_name)==0: # this might happen if pred_dir=='path/to/folder/'
                dataset_name = os.path.basename(os.path.dirname(dir_in))
            dataset_name += ext

            omero_uploader.run(user, pwd, host,upload_id,dataset_name,dir_out)
        print("Done prediction!")

        # print for remote. Format TAG:key:value
        print("REMOTE:dir_out:{}".format(dir_out))
        return dir_out

    elif 'Project' in obj:
        dir_out = os.path.join(dir_out, os.path.split(dir_in)[-1])
        if not os.path.isdir(dir_out):
            os.makedirs(dir_out, exist_ok=True)
        pred.pred_multiple(log, dir_in, dir_out)
        print("Done prediction!")

        # print for remote. Format TAG:key:value
        print("REMOTE:dir_out:{}".format(dir_out))
        return dir_out
    else:
        print("[Error] Type of object unknown {}. It should be 'Dataset' or 'Project'".format(obj))
    
if __name__=='__main__':

    # parser
    parser = argparse.ArgumentParser(description="Prediction with Omero.")
    parser.add_argument('--obj', type=str,
        help="Download object: 'Project:ID' or 'Dataset:ID'")
    parser.add_argument('--target', type=str, default="data/to_pred/",
        help="Directory name to download into")
    parser.add_argument("--log", type=str, default="logs/unet_nucleus",
        help="Path of the builder directory")
    parser.add_argument("--dir_out", type=str, default="data/pred/",
        help="Path to the output prediction directory")
    parser.add_argument('--hostname', type=str, default=None,
        help="(optional) Host name for Omero server. If not mentioned use the CLI.")
    parser.add_argument('--username', type=str, default=None,
        help="(optional) User name for Omero server")
    parser.add_argument('--password', type=str, default=None,
        help="(optional) Password for Omero server")
    parser.add_argument('--upload_id', type=int, default=None,
        help="(optional) Id of Omero Project in which to upload the dataset. Only works with Omero Project Id and folder of images.")
    # parser.add_argument("-e", "--eval_only", default=False,  action='store_true', dest='eval_only',
    #     help="Do only the evaluation and skip the prediction (predictions must have been done already.)") 
    args = parser.parse_args()

    run(
        obj=args.obj,
        target=args.target,
        log=args.log,
        dir_out=args.dir_out,
        host=args.hostname,
        user=args.username,
        pwd=args.password,
        upload_id=args.upload_id,
    )


#---------------------------------------------------------------------------
# Dataset preparation to fasten the training
#   -normalization
#   -expand dims and one_hot encoding
#   -saving to tif file
#---------------------------------------------------------------------------

from sys import platform 
import numpy as np
import os 
import pickle # for foreground storage
from tqdm import tqdm
import argparse
import tifffile
import pandas as pd

from biom3d.auto_config import auto_config, data_fingerprint, get_aug_patch
from biom3d.utils import adaptive_imread, one_hot_fast, resize_3d, save_python_config

np.random.seed(42)

#---------------------------------------------------------------------------
# Define the CSV file for KFold split

def hold_out(df, ratio=0.1, seed=42):
    """
    Select a set of element from the first column of df.
    The size of the set is len(set)*ratio.
    It is randomly selected.
    The results are stored in a new column in df called 'hold_out'.
    The results is a 0/1 list: 1=selected, 0=not selected
    
    Args:
        df: pd.DataFrame
        ratio: float in [0,1]
        seed: np.random.seed initialisation
    Return:
        df: pd.DataFrame
    """
    np.random.seed(seed)
    l = np.array(df.iloc[:,0])
    
    # shuffle the list 
    permut = np.random.permutation(len(l))
    inv_permut = np.argsort(permut)
    # shuffled = l[permut] # shuffled contains the suffled list
    
    # split the shuffled list
    split = int(len(l)*ratio)
    indices = np.array([1]*split+[0]*(len(l)-split))
    
    # unpermut the list of indice to get back the original
    sort_indices = indices[inv_permut]
    
    # add columns
    df['hold_out'] = sort_indices
    return df

def strat_kfold(df, k=4, seed=43):
    """
    Stratified Kfold. 
    Same as kfold but pay attention to balance the train and test sets.
    df must contains a column named 'hold_out'.
    """
    np.random.seed(seed)
    l = np.array(df.iloc[:,0])
    
    holds_out = np.array(df['hold_out'])
    indices_all = np.arange(len(l))
    
    # retrieve train/test indices 
    indices_test = indices_all[holds_out==1]
    indices_train = indices_all[holds_out==0]
    
    # split the list in k folds and shuffle it 
    def split_indices(l):
        kfold_size = len(l)//k
        indices = []
        for i in range(k):
            indices += [i]*kfold_size
        # the remaining indices are randomly assigned
        if len(l[len(indices):])>0:
            for i in range(len(l[len(indices):])):
                alea = np.random.randint(0,k)
                indices += [alea]
        indices = np.array(indices)
        assert len(indices) == len(l)
        np.random.shuffle(indices) # shuffle the indices
        return indices
    
    folds_train = split_indices(indices_train)
    folds_test = split_indices(indices_test)
    
    # merge folds at the right place
    merge = np.zeros_like(l)
    merge[holds_out==1] = folds_test
    merge[holds_out==0] = folds_train
    
    # add the column to the DataFrame
    df['fold'] = merge
    return df

def generate_kfold_csv(filenames, csv_path, hold_out_rate=0., kfold=5, seed=42):
    """From a list of filenames create a CSV containing three columns:
    - filename: image filename
    - hold: 0 or 1, whether to consider this image as test set
    - fold: 0, 1, ..., K, each corresponds to the validation set images

    Parameters
    ----------
    filenames : list of str
        List of the filenames (not the absolute path).
    csv_path : str
        Path of the output csv file.
    hold_out_rate : float, default=0.
        Float between 0 and 1, rate with which the test split will be selected. 
    kfold : int, default=5
        Number of fold that will be defined
    seed : int, default=42
        Random seed for numpy.random
    """
    df = pd.DataFrame(filenames, columns=['filename'])
    df = hold_out(df, ratio=hold_out_rate, seed=seed)
    df = strat_kfold(df, k=kfold, seed=seed)
    df.to_csv(csv_path, index=False)

#---------------------------------------------------------------------------
# 3D segmentation preprocessing

def resize_img_msk(img, output_shape, msk=None):
    new_img = resize_3d(img, output_shape, order=3)
    if msk is not None:
        new_msk = resize_3d(msk, output_shape, is_msk=True, order=1)
        return new_img, new_msk
    else: 
        return new_img

def get_resample_shape(input_shape, spacing, median_spacing):
    input_shape = np.array(input_shape)
    spacing = np.array(spacing)
    median_spacing = np.array(median_spacing)
    if len(input_shape)==4:
        input_shape=input_shape[1:]
    return np.round(((spacing/median_spacing)[::-1]*input_shape)).astype(int)

def sanity_check(msk, num_classes=None):
    """Check if the mask is correctly annotated.
    """
    uni = np.sort(np.unique(msk))
    if num_classes is None:
        num_classes = len(uni)
        
    assert type(num_classes)==int
    assert num_classes >= 2
    
    if len(msk.shape)==4:
        # if we have 4 dimensions in the mask, we consider it one-hot encoded
        # and thus we perform a sanity check for each channel
        for i in range(msk.shape[0]):
            sanity_check(msk[i], num_classes=2)
            
    cls = np.arange(num_classes)
    if np.array_equal(uni,cls):
        # the mask is correctly annotated
        return msk
    else:
        # there is something wrong with the annotations
        # depending on the case we make automatic adjustments
        # or we through an error message
        print("[Warning] There is something abnormal with the annotations. Each voxel value must be in range {} but is in range {}.".format(cls, uni))
        if num_classes==2:
            # thr = np.maximum(msk.min(),0)
            uni2, counts = np.unique(msk,return_counts=True)
            thr = uni2[np.argmax(counts)]
            print("[Warning] All values equal to the most frequent value ({}) will be set to zero.".format(thr))
            # then we apply a threshold to the data
            # for instance: unique [2,127,232] becomes [0,1], 0 being 2 and 1 being 127 and 232
            return (msk != thr).astype(np.uint8)
        elif np.all(np.isin(uni, cls)):
            # then one label is missing in the current mask... but it should work
            print("[Warning] One or more labels are missing.")
            return msk
        elif len(uni)==num_classes:
            # then we re-annotate the unique values in the mask
            # for instance: unique [2,127,232] becomes [0,1,2]
            print("[Warning] Annotation are wrong in the mask, we will re-annotate the mask.")
            new_msk = np.zeros(msk.shape, dtype=msk.dtype)
            for i,c in enumerate(uni):
                new_msk[msk == c] = i
            return new_msk
        else:
            # case like [2,18,128,254] where the number of classes should be 3 are impossible to decide...
            print("[Error] There is an error in the labels that could not be solved automatically.")
            raise RuntimeError

def seg_preprocessor(
    img, 
    img_meta,
    msk=None,
    num_classes=None,
    use_one_hot = False,
    remove_bg = False, 
    median_spacing=[],
    clipping_bounds=[],
    intensity_moments=[],
    channel_axis=0,
    num_channels=1,
    ):
    """Segmentation pre-processing.
    """
    do_msk = msk is not None

    # read image and mask
    spacing = None if not 'spacing' in img_meta.keys() else img_meta['spacing']

    if do_msk: 
        # sanity check
        msk = sanity_check(msk, num_classes)

    # expand image dim
    if len(img.shape)==3:
        # keep the input shape, used for preprocessing before prediction
        original_shape = img.shape
        img = np.expand_dims(img, 0)
    elif len(img.shape)==4:
        # we consider as the channel dimension, the smallest dimension
        # if it is the last dim, then we move it to the first
        # the size of other dimensions of the image should be bigger than the channel dim.
        if np.argmin(img.shape)==channel_axis and img.shape[channel_axis]==num_channels:
            img = np.swapaxes(img, 0, channel_axis)
        else:
            print("[Error] Invalid image shape:", img.shape)
        
        # keep the input shape, used for preprocessing before prediction
        original_shape = img.shape
    else:
        print("[Error] Invalid image shape for 3D image:", img.shape)
        raise ValueError

    assert img.shape[0]==num_channels, "[Error] Invalid image shape {}. Expected to have {} numbers of channel at {} channel axis.".format(img.shape, num_channels, channel_axis)

    # one hot encoding for the mask if needed
    if do_msk and len(msk.shape)!=4: 
        if use_one_hot:
            msk = one_hot_fast(msk, num_classes)
            if remove_bg:
                msk = msk[1:]
        else:
            msk = np.expand_dims(msk, 0)
    elif do_msk and len(msk.shape)==4:
        # normalize each channel
        msk = (msk > msk.min()).astype(np.uint8)

    assert len(img.shape)==4
    if do_msk: assert len(msk.shape)==4

    # clip img
    if len(clipping_bounds)>0:
        img = np.clip(img, clipping_bounds[0], clipping_bounds[1])

    # normalize the image and msk
    # z-score normalization for the image
    if len(intensity_moments)>0:
        img = (img-intensity_moments[0])/intensity_moments[1]
    else:
        img = (img-img.mean())/img.std()
    
    # enhance contrast
    # img = exposure.equalize_hist(img)

    # range image in [-1, 1]
    # img = (img - img.min())/(img.max()-img.min()) * 2 - 1

    # resample the image and mask if needed
    if len(median_spacing)>0 and spacing is not None and len(spacing)>0:
        output_shape = get_resample_shape(img.shape, spacing, median_spacing)
        if do_msk:
            # img, msk = resample_img_msk(img, msk, spacing, median_spacing)
            img, msk = resize_img_msk(img, msk=msk, output_shape=output_shape)
        else:
            # img = resample_with_spacing(img, spacing, median_spacing, order=3)
            img = resize_3d(img, output_shape)

    # set image type
    img = img.astype(np.float32)
    if do_msk: msk = msk.astype(np.uint16)
    
    # foreground computation
    if do_msk:
        fg={}
        if use_one_hot: start = 0 if remove_bg else 1
        else: start = 1
        for i in range(start,len(msk) if use_one_hot else msk.max()+1):
            fgi = np.argwhere(msk[i] == 1) if use_one_hot else np.argwhere(msk[0] == i)
            if len(fgi)>0:
                num_samples = min(len(fgi), 10000)
                fgi_idx = np.random.choice(np.arange(len(fgi)), size=num_samples, replace=False)
                fgi = fgi[fgi_idx,:]
            else:
                fgi = []
            fg[i] = fgi

        if len(fg)==0:
            print("[Warning] Empty foreground!")

    # return
    if do_msk:
        return img, msk, fg 
    else:
        img_meta["original_shape"] = original_shape
        return img, img_meta

#---------------------------------------------------------------------------
# 3D segmentation preprocessing
# Nifti convertion (Medical segmentation decathlon)
# normalization: z-score
# resampling
# intensity normalization
# one_hot encoding

class Preprocessing:
    """A helper class to transform nifti (.nii.gz) and Tiff (.tif or .tiff) images to .npy format and to normalize them.

    Parameters
    ----------
    img_dir : str
        Path to the input image folder
    img_outdir : str
        Path to the output image folder
    msk_dir : str, optional
        Path to the input mask folder
    msk_outdir : str, optional
        Path to the output mask folder
    fg_outdir : str, optional
        Foreground location, eventually later used by the dataloader.
    num_classes : int, optional
        Number of classes (channel) in the masks. Required by the 
    remove_bg : bool, default=True
        Whether to remove the background in the one-hot encoded mask. Remove the background is done when training with sigmoid activations instead of softmax.
    median_size : list, optional
        Median size of the image dataset. Is used to check the channel axis.
    median_spacing : list, optional
        A list of length 3 containing the median spacing of the input images. Median_spacing must not be transposed: for example, median_spacing might be [0.8, 0.8, 2.5] if median shape of the training image is [40,224,224].
    clipping_bounds : list, optional
        A list of length 2 containing the intensity clipping boundary. In nnUNet implementation it corresponds to the 0.5 an 99.5 percentile of the intensities of the voxels of the training images located inside the masks regions.
    intensity_moments : list, optional
        Mean and variance of the intensity of the images voxels in the masks regions. These values are used to normalize the image. 
    use_tif : bool, default=True
        Use tif format to save the preprocessed images instead of npy format.
    split_rate_for_single_img : float, default=0.2
        If a single image is present in image/mask folders, then the image/mask are split in 2 portions of size split_rate_for_single_img*largest_dimension for validation and split_rate_for_single_img*(1-largest_dimension) for training.
    num_kfolds : int, default=5
        Number of K-fold for cross validation.
    """
    def __init__(
        self,
        img_dir,
        img_outdir = None,
        msk_dir = None, # if None, only images are preprocesses not the masks
        msk_outdir = None,
        fg_outdir = None, # foreground location, eventually used by the dataloader
        num_classes = None, # just for debug when empty masks are provided
        use_one_hot = False,
        remove_bg = False, # keep the background in labels 
        median_size = [],
        median_spacing=[],
        clipping_bounds=[],
        intensity_moments=[],
        use_tif=False, # use tif instead of npy 
        split_rate_for_single_img=0.25,
        num_kfolds=5,
        ):
        assert img_dir!='', "[Error] img_dir must not be empty."

        # fix bug path/folder/ to path/folder
        if os.path.basename(img_dir)=='':
            img_dir = os.path.dirname(img_dir)
        if msk_dir is not None and os.path.basename(msk_dir)=='':
            msk_dir = os.path.dirname(msk_dir)
        
        self.img_dir=img_dir
        self.msk_dir=msk_dir
        self.img_fnames=sorted(os.listdir(self.img_dir))

        if img_outdir is None: # name the out dir the same way as the input and add the _out suffix
            img_outdir = img_dir+'_out'
            print("Image output path:", img_outdir)
        if msk_dir is not None and msk_outdir is None:
            msk_outdir = msk_dir+'_out'
            print("Mask output path:", msk_outdir)
            if fg_outdir is None:
                # get parent directory of mask dir
                fg_outdir = os.path.join(os.path.dirname(msk_dir), 'fg_out')
                print("Foreground output path:", fg_outdir)

        self.img_outdir=img_outdir 
        self.msk_outdir=msk_outdir
        self.fg_outdir =fg_outdir

        # create output directory if needed
        if not os.path.exists(self.img_outdir):
            os.makedirs(self.img_outdir, exist_ok=True)
        if msk_dir is not None and not os.path.exists(self.msk_outdir):
            os.makedirs(self.msk_outdir, exist_ok=True)
        if msk_dir is not None and not os.path.exists(self.fg_outdir):
            os.makedirs(self.fg_outdir, exist_ok=True)

        # create csv along with the img folder
        self.csv_path = os.path.join(os.path.dirname(img_dir), 'folds.csv')

        self.num_classes = num_classes

        self.remove_bg = remove_bg

        # median size serves to determine the number of channel
        # and the channel axis
        self.median_size = np.array(median_size)

        self.num_channels = 1
        self.channel_axis = 0

        # if the 3D image has 4 dimensions then there is a channel dimension.
        if len(self.median_size)==4:
            # the channel dimension is consider to be the smallest dimension
            # this could cause problem in case where there are more z than c for instance...
            self.num_channels = np.min(median_size)
            self.channel_axis = np.argmin(self.median_size)
            if self.channel_axis != 0:
                print("[Warning] 4 dimensions detected and channel axis is {}. All image dimensions will be swapped.".format(self.channel_axis))
            self.median_size[[0,self.channel_axis]] = self.median_size[[self.channel_axis,0]]
            self.median_size = self.median_size[1:]

        self.median_spacing = np.array(median_spacing)
        self.clipping_bounds = np.array(clipping_bounds)
        self.intensity_moments = intensity_moments
        self.use_tif = use_tif

        self.split_rate_for_single_img = split_rate_for_single_img

        self.use_one_hot = use_one_hot

        self.num_kfolds = num_kfolds
        if self.num_kfolds * 2 > len(self.img_fnames):
            self.num_kfolds = max(len(self.img_fnames) // 2, 2)
            print("[Warning] The number of images {} is smaller than twice the number of folds {}. The number of folds will be reduced to {}.".format(len(self.img_fnames), num_kfolds * 2, self.num_kfolds))
        
    def _split_single(self):
        """
        if there is only a single image/mask in each folder, then split them both in two portions with self.split_rate_for_single_img
        """
        # set image and mask name
        img_fname = self.img_fnames[0]
        img_path = os.path.join(self.img_dir, img_fname)
        msk_path = os.path.join(self.msk_dir, img_fname) # mask must be present

        # read image and mask
        img, metadata = adaptive_imread(img_path)
        msk = adaptive_imread(msk_path)[0]

        # determine the slicing indices to crop an image along its maximum dimension
        idx = lambda start,end,shape: tuple(slice(s) if s!=max(shape) else slice(start,end) for s in shape)

        # slicing indices of the image
        # validation is cropped along its largest dimension in the interval [0, self.split_rate_for_single_img*largest_dim]
        # training is cropped along its largest dimension in the interval [self.split_rate_for_single_img*largest_dim, largest_dim]
        s = max(img.shape)
        val_img_idx = idx(start=0, end=int(np.floor(self.split_rate_for_single_img*s)), shape=img.shape)
        train_img_idx = idx(start=int(np.floor(self.split_rate_for_single_img*s)), end=s, shape=img.shape)

        # idem for the mask indices
        s = max(msk.shape)
        val_msk_idx = idx(start=0, end=int(np.floor(self.split_rate_for_single_img*s)), shape=msk.shape)
        train_msk_idx = idx(start=int(np.floor(self.split_rate_for_single_img*s)), end=s, shape=msk.shape)

        # crop the images and masks
        val_img = img[val_img_idx]
        train_img = img[train_img_idx]
        val_msk = msk[val_msk_idx]
        train_msk = msk[train_msk_idx]

        # save the images and masks 
        # validation names start with a 0
        # training names start with a 1

        # validation
        val_img_name = "0_"+os.path.basename(img_path).split('.')[0]
        val_img_name += '.tif' if self.use_tif else '.npy'

        val_img_path = os.path.join(self.img_outdir, val_img_name)
        if self.use_tif:
            tifffile.imwrite(val_img_path, val_img, compression=('zlib', 1))
        else:
            np.save(val_img_path, val_img)

        val_msk_path = os.path.join(self.msk_outdir, val_img_name)
        if self.use_tif:
            tifffile.imwrite(val_msk_path, val_msk, compression=('zlib', 1))
        else:
            np.save(val_msk_path, val_msk)

        # training save
        train_img_name = "1_"+os.path.basename(img_path).split('.')[0]
        train_img_name += '.tif' if self.use_tif else '.npy'

        train_img_path = os.path.join(self.img_outdir, train_img_name)
        if self.use_tif:
            tifffile.imwrite(train_img_path, train_img, compression=('zlib', 1))
        else:
            np.save(train_img_path, train_img)

        train_msk_path = os.path.join(self.msk_outdir, train_img_name)
        if self.use_tif:
            tifffile.imwrite(train_msk_path, train_msk, compression=('zlib', 1))
        else:
            np.save(train_msk_path, train_msk)

        # replace self.img_fnames and self.img_dir/self.msk_dir
        self.img_fnames = os.listdir(self.img_outdir)
        self.img_dir = self.img_outdir
        self.msk_dir = self.msk_outdir

        # generate the csv file
        df = pd.DataFrame([train_img_name, val_img_name], columns=['filename'])
        df['hold_out'] = [0,0]
        df['fold'] = [1,0]
        df.to_csv(self.csv_path, index=False)
        return metadata
    
    def run(self, debug=False):
        """Start the preprocessing.

        Parameters
        ----------
        debug : boolean, default=False
            Whether to display image filenames while preprocessing.
        """
        print("Preprocessing...")
        # if there is only a single image/mask, then split them both in two portions
        image_was_split = False
        if len(self.img_fnames)==1 and self.msk_dir is not None:
            print("Single image found per folder. Split the images...")
            split_meta = self._split_single()
            image_was_split = True
        
        if debug: ran = range(len(self.img_fnames))
        else: ran = tqdm(range(len(self.img_fnames)))
        for i in ran:
            # set image and mask name
            img_fname = self.img_fnames[i]
            img_path = os.path.join(self.img_dir, img_fname)
            if self.msk_dir is not None: msk_path = os.path.join(self.msk_dir, img_fname)

            # print image name if debug mode
            if debug: 
                print("[{}/{}] Preprocessing:".format(i,len(self.img_fnames)),img_path)

            img,img_meta = adaptive_imread(img_path)
            if self.msk_dir is not None:
                msk, _ = adaptive_imread(msk_path)
                img, msk, fg = seg_preprocessor(
                    img                 =img, 
                    img_meta            =img_meta if not image_was_split else split_meta,
                    msk                 =msk,
                    num_classes         =self.num_classes,
                    use_one_hot         =self.use_one_hot,
                    remove_bg           =self.remove_bg, 
                    median_spacing      =self.median_spacing,
                    clipping_bounds     =self.clipping_bounds,
                    intensity_moments   =self.intensity_moments,
                    channel_axis        =self.channel_axis,
                    num_channels        =self.num_channels,
                    )
            else:
                img, _ = seg_preprocessor(
                    img                 =img, 
                    img_meta            =img_meta,
                    median_spacing      =self.median_spacing,
                    clipping_bounds     =self.clipping_bounds,
                    intensity_moments   =self.intensity_moments,
                    channel_axis        =self.channel_axis,
                    num_channels        =self.num_channels,
                    )

            # sanity check to be sure that all images have the save number of channel
            s = img.shape
            if len(s)==4: # only for images with 4 dimensionalities
                if i==0: self.num_channels = s[0]
                else: assert len(s)==4 and self.num_channels==s[0], "[Error] Not all images have {} channels. Problematic image: {}".format(self.num_channels, img_path)

            # save the image and the mask as tif
            img_fname = os.path.basename(img_path).split('.')[0]
            # save image

            # save image as tif
            if self.use_tif:
                img_out_path = os.path.join(self.img_outdir, img_fname+'.tif')
                tifffile.imwrite(img_out_path, img, compression=('zlib', 1))
                # imsave(img_out_path, img)
                # tifffile.imwrite(img_out_path, img) # no compression --> increased training speed!
            # save image as npy
            else:
                img_out_path = os.path.join(self.img_outdir, img_fname+'.npy')
                np.save(img_out_path, img)

            # save mask
            if self.msk_outdir is not None: 
                # imsave(msk_out_path, msk)
                # save image as tif
                if self.use_tif:
                    msk_out_path = os.path.join(self.msk_outdir, img_fname+'.tif')
                    tifffile.imwrite(msk_out_path, msk, compression=('zlib', 1))
                    # imsave(msk_out_path, msk)
                    # tifffile.imwrite(msk_out_path, msk) # no compression --> increased training speed!
                # save image as npy
                else:
                    msk_out_path = os.path.join(self.msk_outdir, img_fname+'.npy')
                    np.save(msk_out_path, msk)

                # store it in a pickle format
                fg_file = os.path.join(self.fg_outdir, img_fname+'.pkl')
                with open(fg_file, 'wb') as handle:
                    pickle.dump(fg, handle, protocol=pickle.HIGHEST_PROTOCOL)

        # create csv file
        filenames = sorted(os.listdir(self.img_outdir))
        if not image_was_split:
            generate_kfold_csv(filenames, self.csv_path, kfold=self.num_kfolds)

        print("Done preprocessing!")

#---------------------------------------------------------------------------

def auto_config_preprocess(
        img_dir, 
        msk_dir, 
        num_classes, 
        config_dir, 
        base_config, 
        img_outdir=None,
        msk_outdir=None,
        use_one_hot=False,
        ct_norm=False,
        remove_bg=False, 
        use_tif=False,
        desc="unet", 
        max_dim=128,
        num_epochs=1000,
        skip_preprocessing=False,
        no_auto_config=False,
        logs_dir='logs/',
        print_param=False,
        debug=False,
        ):
    """Helper function to do auto-config and preprocessing.
    """
    
    median_size, median_spacing, mean, std, perc_005, perc_995 = data_fingerprint(img_dir, msk_dir if ct_norm else None)
    if not print_param:
        print("Data fingerprint:")
        print("Median size:", median_size)
        print("Median spacing:", median_spacing)
        print("Mean intensity:", mean)
        print("Standard deviation of intensities:", std)
        print("0.5% percentile of intensities:", perc_005)
        print("99.5% percentile of intensities:", perc_995)
        print("")

    if ct_norm:
        if not print_param: print("Computing data fingerprint for CT normalization...")
        clipping_bounds = [perc_005, perc_995]
        intensity_moments = [mean, std]
        if not print_param: print("Done!")
    else:
        # median_size = None
        # median_spacing = []
        # if sum(median_spacing)==len(median_size): # in case spacing all = 1 = default value
        #     median_spacing = []
        clipping_bounds = []
        intensity_moments = []

    p=Preprocessing(
        img_dir=img_dir,
        msk_dir=msk_dir,
        img_outdir=img_outdir,
        msk_outdir=msk_outdir,
        num_classes=num_classes+1,
        use_one_hot=use_one_hot,
        remove_bg=remove_bg,
        use_tif=use_tif,
        median_spacing=median_spacing,
        median_size=median_size,
        clipping_bounds=clipping_bounds,
        intensity_moments=intensity_moments,
    )

    if not skip_preprocessing:
        p.run(debug=debug)

    if not no_auto_config:
        if not print_param: print("Start auto-configuration")
        

        batch, aug_patch, patch, pool = auto_config(
            median=p.median_size,
            img_dir=img_dir if p.median_size is None else None,
            max_dims=(max_dim, max_dim, max_dim),
            max_batch = len(os.listdir(img_dir))//20, # we limit batch to avoid overfitting
            )
        
        # convert path for windows systems before writing them
        if platform=='win32':
            p.img_outdir = p.img_outdir.replace('\\','\\\\')
            p.msk_outdir = p.msk_outdir.replace('\\','\\\\')
            p.fg_outdir = p.fg_outdir.replace('\\','\\\\')
            p.csv_path = p.csv_path.replace('\\','\\\\')

        config_path = save_python_config(
            config_dir=config_dir,
            base_config=base_config,

            # store hyper-parameters in the config file:
            IMG_DIR=p.img_outdir,
            MSK_DIR=p.msk_outdir,
            FG_DIR=p.fg_outdir,
            CSV_DIR=p.csv_path,
            NUM_CLASSES=num_classes,
            NUM_CHANNELS=p.num_channels,
            CHANNEL_AXIS=p.channel_axis,
            BATCH_SIZE=batch,
            AUG_PATCH_SIZE=aug_patch,
            PATCH_SIZE=patch,
            NUM_POOLS=pool,
            MEDIAN_SPACING=median_spacing,
            CLIPPING_BOUNDS=clipping_bounds,
            INTENSITY_MOMENTS=intensity_moments,
            DESC=desc,
            NB_EPOCHS=num_epochs,
            LOG_DIR=logs_dir,
        )

        if not print_param: print("Auto-config done! Configuration saved in: ", config_path)
        if print_param:
            print(batch)
            print(patch)
            print(aug_patch)
            print(pool)
            print(config_path)

        return config_path

if __name__=='__main__':

    parser = argparse.ArgumentParser(description="Dataset preprocessing for training purpose.")
    parser.add_argument("--img_dir", type=str,
        help="Path of the images directory")
    parser.add_argument("--msk_dir", type=str, default=None,
        help="(default=None) Path to the masks/labels directory")
    parser.add_argument("--img_outdir", type=str, default=None,
        help="(default=None) Path to the directory of the preprocessed images")
    parser.add_argument("--msk_outdir", type=str, default=None,
        help="(default=None) Path to the directory of the preprocessed masks/labels")
    parser.add_argument("--num_classes", type=int, default=1,
        help="(default=1) Number of classes (types of objects) in the dataset. The background is not included.")
    parser.add_argument("--max_dim", type=int, default=128,
        help="(default=128) max_dim^3 determines the maximum size of patch for auto-config.")
    parser.add_argument("--num_epochs", type=int, default=1000,
        help="(default=1000) Number of epochs for the training.")
    parser.add_argument("--config_dir", type=str, default='configs/',
        help="(default=\'configs/\') Configuration folder to save the auto-configuration.")
    parser.add_argument("--logs_dir", type=str, default='logs/',
        help="(default=\'logs/\') Builder output folder to save the model.")
    parser.add_argument("--base_config", type=str, default=None,
        help="(default=None) Optional. Path to an existing configuration file which will be updated with the preprocessed values.")
    parser.add_argument("--desc", type=str, default='unet_default',
        help="(default=unet_default) Optional. A name used to describe the model.")
    parser.add_argument("--use_tif", default=False,  action='store_true', dest='use_tif',
        help="(default=False) Whether to use tif format to save the preprocessed images instead of npy format. Tif files are easily readable with viewers such as Napari and takes fewer disk space but are slower to load and may slow down the training process.") 
    parser.add_argument("--use_one_hot", default=False,  action='store_true', dest='use_one_hot',
        help="(default=False) Whether to use one hot encoding of the mask. Can slow down the training.") 
    parser.add_argument("--remove_bg", default=False,  action='store_true', dest='remove_bg',
        help="(default=False) If use one hot, remove the background in masks. Remove the bg to use with sigmoid activation maps (not softmax).") 
    parser.add_argument("--no_auto_config", default=False,  action='store_true', dest='no_auto_config',
        help="(default=False) For debugging, deactivate auto-configuration.") 
    parser.add_argument("--ct_norm", default=False,  action='store_true', dest='ct_norm',
        help="(default=False) Whether to use CT-Scan normalization routine (cf. nnUNet).") 
    parser.add_argument("--skip_preprocessing", default=False,  action='store_true', dest='skip_preprocessing',
        help="(default=False) Whether to skip the preprocessing. Only for debugging.") 
    parser.add_argument("--remote", default=False,  action='store_true', dest='remote',
        help="(default=False) Whether to print auto-config parameters. Used for remote preprocessing using the GUI.") 
    parser.add_argument("--debug", default=False,  action='store_true', dest='debug',
        help="(default=False) Debug mode. Whether to print all image filenames while preprocessing.") 
    args = parser.parse_args()

    auto_config_preprocess(
        img_dir=args.img_dir, 
        msk_dir=args.msk_dir, 
        num_classes=args.num_classes, 
        config_dir=args.config_dir, 
        base_config=args.base_config, 
        img_outdir=args.img_outdir,
        msk_outdir=args.msk_outdir,
        use_one_hot=args.use_one_hot,
        ct_norm=args.ct_norm,
        remove_bg=args.remove_bg, 
        use_tif=args.use_tif,
        desc=args.desc, 
        max_dim=args.max_dim,
        num_epochs=args.num_epochs,
        skip_preprocessing=args.skip_preprocessing,
        no_auto_config=args.no_auto_config,
        logs_dir=args.logs_dir,
        print_param=args.remote,
        debug=args.debug,
        )

#---------------------------------------------------------------------------




# adtapted from https://gist.github.com/will-moore/a9f90c97b5b6f1a0da277a5179d62c5a 
# documentation: https://downloads.openmicroscopy.org/omero/5.3.1/api/python/omero/omero.plugins.html 

import argparse
import sys
import os

from omero.gateway import BlitzGateway
from omero.cli import cli_login, CLI


from omero.plugins.download import DownloadControl

OBJ_INFO = "obj should be 'Project:ID' or 'Dataset:ID'"

"""
Usage:
python download_pdi.py Project:123 my_project_directory
"""

def download_datasets_cli(datasets, target_dir):
    with cli_login() as cli:
        cli.register("download", DownloadControl, "omero_downloader.py")

        for dataset in datasets:
            print("Downloading Dataset", dataset.id, dataset.name)
            dataset_dir = os.path.join(target_dir, dataset.name)
            os.makedirs(dataset_dir, exist_ok=True)

            for image in dataset.listChildren():
                if image.getFileset() is None:
                    print("No files to download for Image", image.id)
                    continue
                # image_dir = os.path.join(dataset_dir, image.name)
                # If each image is a single file, or are guaranteed not to clash
                # then we don't need image_dir. Could use dataset_dir instead
                # cli.invoke(["download", f'Image:{image.id}', image_dir])
                cli.invoke(["download", f'Image:{image.id}', dataset_dir])


def download_object_cli(cli, obj, target_dir):
    """
    usage:
    with cli_login() as cli:
        download_object(cli, args.obj, args.target)
    """

    conn = BlitzGateway(client_obj=cli._client)
    conn.SERVICE_OPTS.setOmeroGroup(-1)

    try:
        obj_id = int(obj.split(":")[1])
        obj_type = obj.split(":")[0]
    except:
        print(OBJ_INFO)

    parent = conn.getObject(obj_type, obj_id)
    if parent is None:
        print("Not Found:", obj)

    datasets = []

    if obj_type == "Dataset":
        datasets.append(parent)
    elif obj_type == "Project":
        datasets = list(parent.listChildren())
        target_dir = os.path.join(target_dir, parent.getName())
    else:
        print(OBJ_INFO)

    print("Downloading to ", target_dir)

    download_datasets(datasets, target_dir)

    return datasets, target_dir

def download_datasets(conn, datasets, target_dir):

    for dataset in datasets:
        print("Downloading Dataset", dataset.id, dataset.name)
        dc = DownloadControl()
        dataset_dir = os.path.join(target_dir, dataset.name)
        os.makedirs(dataset_dir, exist_ok=True)

        for image in dataset.listChildren():
            if image.getFileset() is None:
                print("No files to download for Image", image.id)
                continue
            # image_dir = os.path.join(dataset_dir, image.name)
            # If each image is a single file, or are guaranteed not to clash
            # then we don't need image_dir. Can use dataset_dir instead
            
            fileset = image.getFileset()
            if fileset is None:
                print('Image has no Fileset')
                continue
            dc.download_fileset(conn, fileset, dataset_dir)

def download_object(username, password, hostname, obj, target_dir):
    conn = BlitzGateway(username=username, passwd=password, host=hostname, port=4064)
    conn.connect()
    try:
        obj_id = int(obj.split(":")[1])
        obj_type = obj.split(":")[0]
    except:
        print(OBJ_INFO)

    parent = conn.getObject(obj_type, obj_id)
    if parent is None:
        print("Not Found:", obj)

    datasets = []

    if obj_type == "Dataset":
        datasets.append(parent)
    elif obj_type == "Project":
        datasets = list(parent.listChildren())
        target_dir = os.path.join(target_dir, parent.getName())
    else:
        print(OBJ_INFO)

    print("Downloading to ", target_dir)

    download_datasets(conn, datasets, target_dir)

    conn.close()

    return datasets, target_dir


def main(argv):
    parser = argparse.ArgumentParser()
    parser.add_argument('--obj',
        help="Download object: 'Project:ID' or 'Dataset:ID'")
    parser.add_argument('--target',
        help="Directory name to download into")
    parser.add_argument('--username',
        help="User name")
    parser.add_argument('--password',
        help="Password")
    parser.add_argument('--hostname',
        help="Host name")
    args = parser.parse_args(argv)

    download_object(args.username, args.password, args.hostname, args.obj, args.target)

if __name__ == '__main__':
    main(sys.argv[1:])


import numpy as np
import argparse

from biom3d.utils import abs_listdir, versus_one, dice

def robust_sort(str_list):
    """Robust sorting of string. Useful for list of paths sorting.
    """
    # max string lenght
    max_len = max(list(len(s) for s in str_list))

    # add zeros in the beginning so that all strings have the same length
    # associate with the original length to the elongated string
    same_len = {'0'*(max_len-len(s))+s:len(s) for s in str_list}

    # sort the dict by key
    sorted_same_len = {k:same_len[k] for k in sorted(same_len)}

    # remove zeros and return
    return [k[max_len-v:] for k,v in sorted_same_len.items()]

def eval(dir_lab, dir_out, num_classes):
    print("Start evaluation")
    paths_lab = [dir_lab, dir_out]
    list_abs = [sorted(abs_listdir(p)) for p in paths_lab]
    assert sum([len(t) for t in list_abs])%len(list_abs)==0, "[Error] Not the same number of labels and predictions! {}".format([len(t) for t in list_abs])

    results = []
    for idx in range(len(list_abs[0])):
        print("Metric computation for:", list_abs[1][idx])
        res = versus_one(
            fct=dice, 
            in_path=list_abs[1][idx], 
            tg_path=list_abs[0][idx], 
            num_classes=num_classes+1, 
            single_class=None)

        print("Metric result:", res)
        results += [res]
        
    print("Evaluation done! Average result:", np.mean(results))

if __name__=='__main__':
    parser = argparse.ArgumentParser(description="Prediction evaluation.")
    parser.add_argument("-p", "--dir_pred", type=str, default=None,
        help="Path to the prediction directory")  
    parser.add_argument("-l", "--dir_lab", type=str, default=None,
        help="Path to the label directory")  
    parser.add_argument("--num_classes", type=int, default=1,
        help="(default=1) Number of classes (types of objects) in the dataset. The background is not included.")
    args = parser.parse_args()

    eval(args.dir_pred, args.dir_lab, args.num_classes)



#---------------------------------------------------------------------------
# model trainers
# TODO: re-structure with classes maybe?
#---------------------------------------------------------------------------

import torch 
import torchio as tio
from tqdm import tqdm
from time import time 
import math 
import sys 

#---------------------------------------------------------------------------
# model trainers for segmentation

def seg_train(
    dataloader, 
    scaler,
    model, 
    loss_fn,
    metrics, 
    optimizer, 
    callbacks, 
    epoch = None, # required by deep supervision
    use_deep_supervision=False):

    model.train()
    
    if torch.cuda.is_available():
        torch.cuda.synchronize()
    t_start_epoch = time()
    print("[time] start epoch")

    for batch, (X, y) in enumerate(dataloader):
        
        callbacks.on_batch_begin(batch)
        if torch.cuda.is_available():
            X, y = X.cuda(), y.cuda()

            # print("batch:",batch, "memory reserved:", torch.cuda.memory_reserved(0), "allocated memory", torch.cuda.memory_allocated(0), "free memory:", torch.cuda.memory_reserved(0)-torch.cuda.memory_allocated(0), "percentage of free memory:", torch.cuda.memory_allocated(0)/torch.cuda.max_memory_allocated(0))

            torch.cuda.synchronize()
        t_data_loading = time()

        if t_data_loading-t_start_epoch > 1:
            print("SLOW!", batch, "[time] data loading:", t_data_loading-t_start_epoch)

        # Compute prediction error

        # with CUDA
        if torch.cuda.is_available():
            with torch.cuda.amp.autocast(scaler is not None):
                pred = model(X); del X
                loss = loss_fn(pred, y)
                with torch.no_grad():
                    if use_deep_supervision:
                        for m in metrics: m(pred[-1],y)
                    else: 
                        for m in metrics: m(pred,y)
        
        # with CPU
        else:
            pred = model(X); del X
            loss = loss_fn(pred, y)
            with torch.no_grad():
                if use_deep_supervision:
                    for m in metrics: m(pred[-1],y)
                else: 
                    for m in metrics: m(pred,y)

        # Backpropagation
        optimizer.zero_grad() # set gradient to zero, why is that needed?

        if scaler is not None:
            scaler.scale(loss).backward() # compute gradient

            scaler.unscale_(optimizer)
            torch.nn.utils.clip_grad_norm_(model.parameters(), 12)

            scaler.step(optimizer) # apply gradient
            scaler.update()
        else: 
            loss.backward()
            torch.nn.utils.clip_grad_norm_(model.parameters(), 12)
            optimizer.step()

        del loss, pred, y 

        callbacks.on_batch_end(batch)

        if torch.cuda.is_available():
            torch.cuda.synchronize()
        t_start_epoch = time()
    
    if torch.cuda.is_available():
        torch.cuda.empty_cache()

def seg_validate(
    dataloader,
    model,
    loss_fn,
    metrics,
    use_fp16,
    use_deep_supervision=False):
    for m in [loss_fn]+metrics: m.reset() # reset metrics
    model.eval() # set the module in evaluation mode (only useful for dropout or batchnorm like layers)
    with torch.no_grad(): # set all the requires_grad flags to zeros
        for X, y in dataloader:

            # with CUDA
            if torch.cuda.is_available():
                X, y = X.cuda(), y.cuda()
                with torch.cuda.amp.autocast(use_fp16):
                    pred=model(X)
                    del X
                    loss_fn(pred, y)
                    loss_fn.update()
                    for m in metrics:
                        if use_deep_supervision:
                            m(pred[-1],y)
                        else:
                            m(pred, y)
                        m.update()
                del pred, y
            
            # with CPU
            else:
                pred=model(X)
                del X
                loss_fn(pred, y)
                loss_fn.update()
                for m in metrics:
                    if use_deep_supervision:
                        m(pred[-1],y)
                    else:
                        m(pred, y)
                    m.update()
                del pred, y
                
    if torch.cuda.is_available():
        torch.cuda.empty_cache()
    template = "val error: avg loss {:.3f}".format(loss_fn.avg.item())
    for m in metrics: template += ", " + str(m)
    print(template)

#---------------------------------------------------------------------------
# model trainers for segmentation with patches 

def seg_patch_validate(dataloader, model, loss_fn, metrics):
    print("Start validation...")
    for m in [loss_fn]+metrics: m.reset() # reset metrics
    model.eval() # set the module in evaluation mode (only useful for dropout or batchnorm like layers)
    with torch.no_grad(): # set all the requires_grad flags to zeros
        for it in tqdm(dataloader):
            # pred_aggr = tio.inference.GridAggregator(it, overlap_mode='average')
            # y_aggr = tio.inference.GridAggregator(it, overlap_mode='average')
            patch_loader = torch.utils.data.DataLoader(it, batch_size=dataloader.batch_size)
            for patch in patch_loader:
                X = patch['img'][tio.DATA]
                y = patch['msk'][tio.DATA]
                if torch.cuda.is_available():
                    X, y = X.cuda(), y.cuda()
                pred=model(X).detach()

                # pred_aggr.add_batch(pred, patch[tio.LOCATION])
                # y_aggr.add_batch(y, patch[tio.LOCATION])
            
            # pred = pred_aggr.get_output_tensor()
            # y = y_aggr.get_output_tensor()
                loss_fn(pred, y)
                loss_fn.update()
                for m in metrics:
                    m(pred, y)
                    m.update()

    template = "val error: avg loss {:.3f}".format(loss_fn.avg.item())
    for m in metrics: template += ", " + str(m)
    print(template)

def seg_patch_train(
    dataloader, 
    model, 
    loss_fn,
    metrics, 
    optimizer, 
    callbacks, 
    epoch = None, # required by deep supervision
    use_deep_supervision=False):

    model.train()
    for batch, queue in enumerate(dataloader):
        patch_loader = torch.utils.data.DataLoader(queue, 
            batch_size  =dataloader.batch_size,  
            drop_last   =False, 
            shuffle     =False, 
            num_workers =0, 
            pin_memory  =True)
        for patch, (X, y) in enumerate(patch_loader):
            callbacks.on_batch_begin(batch * len(patch_loader) + patch)

            if torch.cuda.is_available():
                X, y = X.cuda(), y.cuda()

            # Compute prediction error
            pred = model(X)
            if use_deep_supervision:
                loss = loss_fn(pred, y, epoch)
                for m in metrics: m(pred[-1].detach(),y.detach(), epoch)
            else: 
                loss = loss_fn(pred, y)
                for m in metrics: m(pred.detach(),y.detach())

            # Backpropagation
            # for param in optimizer.parameters():
            #     param.grad = None
            optimizer.zero_grad() # set gradient to zero, why is that needed?
            loss.backward() # compute gradient
            optimizer.step() # apply gradient
            loss.detach()

            callbacks.on_batch_end(batch * len(patch_loader) + patch)

#---------------------------------------------------------------------------